# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import datetime
import functools
from abc import ABC
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from disnake.app_commands import ApplicationCommand
from disnake.enums import ApplicationCommandType
from disnake.flags import ApplicationInstallTypes, InteractionContextTypes
from disnake.permissions import Permissions
from disnake.utils import _generated, _overload_with_permissions, async_all, maybe_coroutine

from .cooldowns import BucketType, CooldownMapping, MaxConcurrency
from .errors import CheckFailure, CommandError, CommandInvokeError, CommandOnCooldown

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec, Self

    from disnake.interactions import ApplicationCommandInteraction

    from ._types import AppCheck, Coro, Error, Hook
    from .cog import Cog
    from .interaction_bot_base import InteractionBotBase

    ApplicationCommandInteractionT = TypeVar(
        "ApplicationCommandInteractionT", bound=ApplicationCommandInteraction, covariant=True
    )

    P = ParamSpec("P")

    CommandCallback = Callable[..., Coro[Any]]
    InteractionCommandCallback = Union[
        Callable[Concatenate["CogT", ApplicationCommandInteractionT, P], Coro[Any]],
        Callable[Concatenate[ApplicationCommandInteractionT, P], Coro[Any]],
    ]


__all__ = (
    "InvokableApplicationCommand",
    "default_member_permissions",
    "install_types",
    "contexts",
)


T = TypeVar("T")
AppCommandT = TypeVar("AppCommandT", bound="InvokableApplicationCommand")
CogT = TypeVar("CogT", bound="Cog")
HookT = TypeVar("HookT", bound="Hook")
ErrorT = TypeVar("ErrorT", bound="Error")


def _get_overridden_method(method):
    return getattr(method.__func__, "__cog_special_method__", method)


def wrap_callback(coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        try:
            ret = await coro(*args, **kwargs)
        except CommandError:
            raise
        except asyncio.CancelledError:
            return
        except Exception as exc:
            raise CommandInvokeError(exc) from exc
        return ret

    return wrapped


class InvokableApplicationCommand(ABC):
    """A base class that implements the protocol for a bot application command.

    These are not created manually, instead they are created via the
    decorator or functional interface.

    The following classes implement this ABC:

    - :class:`~.InvokableSlashCommand`
    - :class:`~.InvokableMessageCommand`
    - :class:`~.InvokableUserCommand`

    Attributes
    ----------
    name: :class:`str`
        The name of the command.
    qualified_name: :class:`str`
        The full command name, including parent names in the case of slash subcommands or groups.
        For example, the qualified name for ``/one two three`` would be ``one two three``.
    body: :class:`.ApplicationCommand`
        An object being registered in the API.
    callback: :ref:`coroutine <coroutine>`
        The coroutine that is executed when the command is called.
    cog: Optional[:class:`Cog`]
        The cog that this command belongs to. ``None`` if there isn't one.
    checks: List[Callable[[:class:`.ApplicationCommandInteraction`], :class:`bool`]]
        A list of predicates that verifies if the command could be executed
        with the given :class:`.ApplicationCommandInteraction` as the sole parameter. If an exception
        is necessary to be thrown to signal failure, then one inherited from
        :exc:`.CommandError` should be used. Note that if the checks fail then
        :exc:`.CheckFailure` exception is raised to the :func:`.on_slash_command_error`
        event.
    guild_ids: Optional[Tuple[:class:`int`, ...]]
        The list of IDs of the guilds where the command is synced. ``None`` if this command is global.
    auto_sync: :class:`bool`
        Whether to automatically register the command.
    extras: Dict[:class:`str`, Any]
        A dict of user provided extras to attach to the command.

        .. versionadded:: 2.5
    """

    __original_kwargs__: Dict[str, Any]
    body: ApplicationCommand

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        self = super().__new__(cls)
        # todo: refactor to not require None and change this to be based on the presence of a kwarg
        self.__original_kwargs__ = {k: v for k, v in kwargs.items() if v is not None}
        return self

    def __init__(self, func: CommandCallback, *, name: Optional[str] = None, **kwargs: Any) -> None:
        self.__command_flag__ = None
        self._callback: CommandCallback = func
        self.name: str = name or func.__name__
        self.qualified_name: str = self.name
        # Annotation parser needs this attribute because body doesn't exist at this moment.
        # We will use this attribute later in order to set the allowed contexts.
        self._guild_only: bool = kwargs.get("guild_only", False)
        self.extras: Dict[str, Any] = kwargs.get("extras") or {}

        if not isinstance(self.name, str):
            raise TypeError("Name of a command must be a string.")

        if "default_permission" in kwargs:
            raise TypeError(
                "`default_permission` is deprecated and will always be set to `True`. "
                "See `default_member_permissions` and `contexts` instead."
            )

        # XXX: remove in next major/minor version
        # the parameter was called `integration_types` in earlier stages of the user apps PR.
        # since unknown kwargs unfortunately get silently ignored, at least try to warn users
        # in this specific case
        if "integration_types" in kwargs:
            raise TypeError("`integration_types` has been renamed to `install_types`.")

        try:
            checks = func.__commands_checks__
            checks.reverse()
        except AttributeError:
            checks = kwargs.get("checks", [])

        self.checks: List[AppCheck] = checks

        try:
            cooldown = func.__commands_cooldown__
        except AttributeError:
            cooldown = kwargs.get("cooldown")

        # TODO: Figure out how cooldowns even work with interactions
        if cooldown is None:
            buckets = CooldownMapping(cooldown, BucketType.default)
        elif isinstance(cooldown, CooldownMapping):
            buckets = cooldown
        else:
            raise TypeError("Cooldown must be a an instance of CooldownMapping or None.")
        self._buckets: CooldownMapping = buckets

        try:
            max_concurrency = func.__commands_max_concurrency__
        except AttributeError:
            max_concurrency = kwargs.get("max_concurrency")
        self._max_concurrency: Optional[MaxConcurrency] = max_concurrency

        self.cog: Optional[Cog] = None
        self.guild_ids: Optional[Tuple[int, ...]] = None
        self.auto_sync: bool = True

        self._before_invoke: Optional[Hook] = None
        self._after_invoke: Optional[Hook] = None

    # this should copy all attributes that can be changed after instantiation via decorators
    def _ensure_assignment_on_copy(self, other: AppCommandT) -> AppCommandT:
        other._before_invoke = self._before_invoke
        other._after_invoke = self._after_invoke
        if self.checks != other.checks:
            other.checks = self.checks.copy()
        if self._buckets.valid and not other._buckets.valid:
            other._buckets = self._buckets.copy()
        if self._max_concurrency != other._max_concurrency:
            # _max_concurrency won't be None at this point
            other._max_concurrency = cast(MaxConcurrency, self._max_concurrency).copy()

        if (
            # see https://github.com/DisnakeDev/disnake/pull/678#discussion_r938113624:
            # if these are not equal, then either `self` had a decorator, or `other` got a
            # value from `*_command_attrs`; we only want to copy in the former case
            self.body._default_member_permissions != other.body._default_member_permissions
            and self.body._default_member_permissions is not None
        ):
            other.body._default_member_permissions = self.body._default_member_permissions

        if (
            self.body.install_types != other.body.install_types
            and self.body.install_types is not None  # see above
        ):
            other.body.install_types = ApplicationInstallTypes._from_value(
                self.body.install_types.value
            )

        if (
            self.body.contexts != other.body.contexts
            and self.body.contexts is not None  # see above
        ):
            other.body.contexts = InteractionContextTypes._from_value(self.body.contexts.value)

        try:
            other.on_error = self.on_error
        except AttributeError:
            pass

        return other

    def copy(self: AppCommandT) -> AppCommandT:
        """Create a copy of this application command.

        Returns
        -------
        :class:`InvokableApplicationCommand`
            A new instance of this application command.
        """
        copy = type(self)(self.callback, **self.__original_kwargs__)
        return self._ensure_assignment_on_copy(copy)

    def _update_copy(self: AppCommandT, kwargs: Dict[str, Any]) -> AppCommandT:
        if kwargs:
            kw = kwargs.copy()
            kw.update(self.__original_kwargs__)
            copy = type(self)(self.callback, **kw)
            return self._ensure_assignment_on_copy(copy)
        else:
            return self.copy()

    def _apply_guild_only(self) -> None:
        # If we have a `GuildCommandInteraction` annotation, set `contexts` and `install_types` accordingly.
        # This matches the old pre-user-apps behavior.
        if self._guild_only:
            # n.b. this overwrites any user-specified parameter
            # FIXME(3.0): this should raise if these were set elsewhere (except `*_command_attrs`) already
            self.body.contexts = InteractionContextTypes(guild=True)
            self.body.install_types = ApplicationInstallTypes(guild=True)

    def _apply_defaults(self, bot: InteractionBotBase) -> None:
        self.body._default_install_types = bot._default_install_types
        self.body._default_contexts = bot._default_contexts

    @property
    def dm_permission(self) -> bool:
        """:class:`bool`: Whether this command can be used in DMs."""
        return self.body.dm_permission

    @property
    def default_member_permissions(self) -> Optional[Permissions]:
        """Optional[:class:`.Permissions`]: The default required member permissions for this command.
        A member must have *all* these permissions to be able to invoke the command in a guild.

        This is a default value, the set of users/roles that may invoke this command can be
        overridden by moderators on a guild-specific basis, disregarding this setting.

        If ``None`` is returned, it means everyone can use the command by default.
        If an empty :class:`.Permissions` object is returned (that is, all permissions set to ``False``),
        this means no one can use the command.

        .. versionadded:: 2.5
        """
        return self.body.default_member_permissions

    @property
    def install_types(self) -> Optional[ApplicationInstallTypes]:
        """Optional[:class:`.ApplicationInstallTypes`]: The installation types
        where the command is available. Only available for global commands.

        .. versionadded:: 2.10
        """
        return self.body.install_types

    @property
    def contexts(self) -> Optional[InteractionContextTypes]:
        """Optional[:class:`.InteractionContextTypes`]: The interaction contexts
        where the command can be used. Only available for global commands.

        .. versionadded:: 2.10
        """
        return self.body.contexts

    @property
    def callback(self) -> CommandCallback:
        return self._callback

    def add_check(self, func: AppCheck) -> None:
        """Adds a check to the application command.

        This is the non-decorator interface to :func:`.app_check`.

        Parameters
        ----------
        func
            The function that will be used as a check.
        """
        self.checks.append(func)

    def remove_check(self, func: AppCheck) -> None:
        """Removes a check from the application command.

        This function is idempotent and will not raise an exception
        if the function is not in the command's checks.

        Parameters
        ----------
        func
            The function to remove from the checks.
        """
        try:
            self.checks.remove(func)
        except ValueError:
            pass

    async def __call__(
        self, interaction: ApplicationCommandInteraction, *args: Any, **kwargs: Any
    ) -> Any:
        """|coro|

        Calls the internal callback that the application command holds.

        .. note::

            This bypasses all mechanisms -- including checks, converters,
            invoke hooks, cooldowns, etc. You must take care to pass
            the proper arguments and types to this function.

        """
        if self.cog is not None:
            return await self.callback(self.cog, interaction, *args, **kwargs)
        else:
            return await self.callback(interaction, *args, **kwargs)

    def _prepare_cooldowns(self, inter: ApplicationCommandInteraction) -> None:
        if self._buckets.valid:
            dt = inter.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            bucket = self._buckets.get_bucket(inter, current)  # type: ignore
            if bucket is not None:  # pyright: ignore[reportUnnecessaryComparison]
                retry_after = bucket.update_rate_limit(current)
                if retry_after:
                    raise CommandOnCooldown(bucket, retry_after, self._buckets.type)  # type: ignore

    async def prepare(self, inter: ApplicationCommandInteraction) -> None:
        inter.application_command = self

        if not await self.can_run(inter):
            raise CheckFailure(f"The check functions for command {self.qualified_name!r} failed.")

        if self._max_concurrency is not None:
            await self._max_concurrency.acquire(inter)  # type: ignore

        try:
            self._prepare_cooldowns(inter)
            await self.call_before_hooks(inter)
        except Exception:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)  # type: ignore
            raise

    def is_on_cooldown(self, inter: ApplicationCommandInteraction) -> bool:
        """Checks whether the application command is currently on cooldown.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with the application command currently being invoked.

        Returns
        -------
        :class:`bool`
            A boolean indicating if the application command is on cooldown.
        """
        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(inter)  # type: ignore
        dt = inter.created_at
        current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
        return bucket.get_tokens(current) == 0

    def reset_cooldown(self, inter: ApplicationCommandInteraction) -> None:
        """Resets the cooldown on this application command.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with this application command
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(inter)  # type: ignore
            bucket.reset()

    def get_cooldown_retry_after(self, inter: ApplicationCommandInteraction) -> float:
        """Retrieves the amount of seconds before this application command can be tried again.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with this application command.

        Returns
        -------
        :class:`float`
            The amount of time left on this command's cooldown in seconds.
            If this is ``0.0`` then the command isn't on cooldown.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(inter)  # type: ignore
            dt = inter.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            return bucket.get_retry_after(current)

        return 0.0

    # This method isn't really usable in this class, but it's usable in subclasses.
    async def invoke(self, inter: ApplicationCommandInteraction, *args: Any, **kwargs: Any) -> None:
        await self.prepare(inter)

        try:
            await self(inter, *args, **kwargs)
        except CommandError:
            inter.command_failed = True
            raise
        except asyncio.CancelledError:
            inter.command_failed = True
            return
        except Exception as exc:
            inter.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(inter)  # type: ignore

            await self.call_after_hooks(inter)

    def error(self, coro: ErrorT) -> ErrorT:
        """A decorator that registers a coroutine as a local error handler.

        A local error handler is an error event limited to a single application command.

        Parameters
        ----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the local error handler.

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("The error handler must be a coroutine.")

        self.on_error: Error = coro
        return coro

    def has_error_handler(self) -> bool:
        """Checks whether the application command has an error handler registered."""
        return hasattr(self, "on_error")

    async def _call_local_error_handler(
        self, inter: ApplicationCommandInteraction, error: CommandError
    ) -> Any:
        if not self.has_error_handler():
            return

        injected = wrap_callback(self.on_error)
        if self.cog is not None:
            return await injected(self.cog, inter, error)
        else:
            return await injected(inter, error)

    async def _call_external_error_handlers(
        self, inter: ApplicationCommandInteraction, error: CommandError
    ) -> None:
        """Overridden in subclasses"""
        raise error

    async def dispatch_error(
        self, inter: ApplicationCommandInteraction, error: CommandError
    ) -> None:
        if not await self._call_local_error_handler(inter, error):
            await self._call_external_error_handlers(inter, error)

    async def call_before_hooks(self, inter: ApplicationCommandInteraction) -> None:
        # now that we're done preparing we can call the pre-command hooks
        # first, call the command local hook:
        cog = self.cog
        if self._before_invoke is not None:
            # should be cog if @commands.before_invoke is used
            instance = getattr(self._before_invoke, "__self__", cog)
            # __self__ only exists for methods, not functions
            # however, if @command.before_invoke is used, it will be a function
            if instance:
                await self._before_invoke(instance, inter)  # type: ignore
            else:
                await self._before_invoke(inter)  # type: ignore

        if inter.data.type is ApplicationCommandType.chat_input:
            partial_attr_name = "slash_command"
        elif inter.data.type is ApplicationCommandType.user:
            partial_attr_name = "user_command"
        elif inter.data.type is ApplicationCommandType.message:
            partial_attr_name = "message_command"
        else:
            return

        # call the cog local hook if applicable:
        if cog is not None:
            meth = getattr(cog, f"cog_before_{partial_attr_name}_invoke", None)
            hook = _get_overridden_method(meth)
            if hook is not None:
                await hook(inter)

        # call the bot global hook if necessary
        hook = getattr(inter.bot, f"_before_{partial_attr_name}_invoke", None)
        if hook is not None:
            await hook(inter)

    async def call_after_hooks(self, inter: ApplicationCommandInteraction) -> None:
        cog = self.cog
        if self._after_invoke is not None:
            instance = getattr(self._after_invoke, "__self__", cog)
            if instance:
                await self._after_invoke(instance, inter)  # type: ignore
            else:
                await self._after_invoke(inter)  # type: ignore

        if inter.data.type is ApplicationCommandType.chat_input:
            partial_attr_name = "slash_command"
        elif inter.data.type is ApplicationCommandType.user:
            partial_attr_name = "user_command"
        elif inter.data.type is ApplicationCommandType.message:
            partial_attr_name = "message_command"
        else:
            return

        # call the cog local hook if applicable:
        if cog is not None:
            meth = getattr(cog, f"cog_after_{partial_attr_name}_invoke", None)
            hook = _get_overridden_method(meth)
            if hook is not None:
                await hook(inter)

        # call the bot global hook if necessary
        hook = getattr(inter.bot, f"_after_{partial_attr_name}_invoke", None)
        if hook is not None:
            await hook(inter)

    def before_invoke(self, coro: HookT) -> HookT:
        """A decorator that registers a coroutine as a pre-invoke hook.

        A pre-invoke hook is called directly before the command is called.

        This pre-invoke hook takes a sole parameter, a :class:`.ApplicationCommandInteraction`.

        Parameters
        ----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the pre-invoke hook.

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("The pre-invoke hook must be a coroutine.")

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro: HookT) -> HookT:
        """A decorator that registers a coroutine as a post-invoke hook.

        A post-invoke hook is called directly after the command is called.

        This post-invoke hook takes a sole parameter, a :class:`.ApplicationCommandInteraction`.

        Parameters
        ----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the post-invoke hook.

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("The post-invoke hook must be a coroutine.")

        self._after_invoke = coro
        return coro

    @property
    def cog_name(self) -> Optional[str]:
        """Optional[:class:`str`]: The name of the cog this application command belongs to, if any."""
        return type(self.cog).__cog_name__ if self.cog is not None else None

    async def can_run(self, inter: ApplicationCommandInteraction) -> bool:
        """|coro|

        Checks if the command can be executed by checking all the predicates
        inside the :attr:`~Command.checks` attribute.

        Parameters
        ----------
        inter: :class:`.ApplicationCommandInteraction`
            The interaction with the application command currently being invoked.

        Raises
        ------
        :class:`CommandError`
            Any application command error that was raised during a check call will be propagated
            by this function.

        Returns
        -------
        :class:`bool`
            A boolean indicating if the application command can be invoked.
        """
        original = inter.application_command
        inter.application_command = self

        if inter.data.type is ApplicationCommandType.chat_input:
            partial_attr_name = "slash_command"
        elif inter.data.type is ApplicationCommandType.user:
            partial_attr_name = "user_command"
        elif inter.data.type is ApplicationCommandType.message:
            partial_attr_name = "message_command"
        else:
            return True

        try:
            if inter.bot and not await inter.bot.application_command_can_run(inter):
                raise CheckFailure(
                    f"The global check functions for command {self.qualified_name} failed."
                )

            cog = self.cog
            if cog is not None:
                meth = getattr(cog, f"cog_{partial_attr_name}_check", None)
                local_check = _get_overridden_method(meth)
                if local_check is not None:
                    ret = await maybe_coroutine(local_check, inter)
                    if not ret:
                        return False

            predicates = self.checks
            if not predicates:
                # since we have no checks, then we just return True.
                return True

            return await async_all(predicate(inter) for predicate in predicates)  # type: ignore
        finally:
            inter.application_command = original


@overload
@_generated
def default_member_permissions(
    value: int = 0,
    *,
    add_reactions: bool = ...,
    administrator: bool = ...,
    attach_files: bool = ...,
    ban_members: bool = ...,
    change_nickname: bool = ...,
    connect: bool = ...,
    create_events: bool = ...,
    create_forum_threads: bool = ...,
    create_guild_expressions: bool = ...,
    create_instant_invite: bool = ...,
    create_private_threads: bool = ...,
    create_public_threads: bool = ...,
    deafen_members: bool = ...,
    embed_links: bool = ...,
    external_emojis: bool = ...,
    external_stickers: bool = ...,
    kick_members: bool = ...,
    manage_channels: bool = ...,
    manage_emojis: bool = ...,
    manage_emojis_and_stickers: bool = ...,
    manage_events: bool = ...,
    manage_guild: bool = ...,
    manage_guild_expressions: bool = ...,
    manage_messages: bool = ...,
    manage_nicknames: bool = ...,
    manage_permissions: bool = ...,
    manage_roles: bool = ...,
    manage_threads: bool = ...,
    manage_webhooks: bool = ...,
    mention_everyone: bool = ...,
    moderate_members: bool = ...,
    move_members: bool = ...,
    mute_members: bool = ...,
    priority_speaker: bool = ...,
    read_message_history: bool = ...,
    read_messages: bool = ...,
    request_to_speak: bool = ...,
    send_messages: bool = ...,
    send_messages_in_threads: bool = ...,
    send_polls: bool = ...,
    send_tts_messages: bool = ...,
    send_voice_messages: bool = ...,
    speak: bool = ...,
    start_embedded_activities: bool = ...,
    stream: bool = ...,
    use_application_commands: bool = ...,
    use_embedded_activities: bool = ...,
    use_external_apps: bool = ...,
    use_external_emojis: bool = ...,
    use_external_sounds: bool = ...,
    use_external_stickers: bool = ...,
    use_slash_commands: bool = ...,
    use_soundboard: bool = ...,
    use_voice_activation: bool = ...,
    view_audit_log: bool = ...,
    view_channel: bool = ...,
    view_creator_monetization_analytics: bool = ...,
    view_guild_insights: bool = ...,
) -> Callable[[T], T]:
    ...


@overload
@_generated
def default_member_permissions(
    value: int = 0,
) -> Callable[[T], T]:
    ...


@_overload_with_permissions
def default_member_permissions(value: int = 0, **permissions: bool) -> Callable[[T], T]:
    """A decorator that sets default required member permissions for the application command.
    Unlike :func:`~.has_permissions`, this decorator does not add any checks.
    Instead, it prevents the command from being run by members without *all* required permissions,
    if not overridden by moderators on a guild-specific basis.

    See also the ``default_member_permissions`` parameter for application command decorators.

    .. note::
        This does not work with slash subcommands/groups.

    .. versionadded:: 2.5

    Example
    -------

    This would only allow members with :attr:`~.Permissions.manage_messages` *and*
    :attr:`~.Permissions.view_audit_log` permissions to use the command by default,
    however moderators can override this and allow/disallow specific users and
    roles to use the command in their guilds regardless of this setting.

    .. code-block:: python3

        @bot.slash_command()
        @commands.default_member_permissions(manage_messages=True, view_audit_log=True)
        async def purge(inter, num: int):
            ...

    Parameters
    ----------
    value: :class:`int`
        A raw permission bitfield of an integer representing the required permissions.
        May be used instead of specifying kwargs.
    **permissions: bool
        The required permissions for a command. A member must have *all* these
        permissions to be able to invoke the command.
        Setting a permission to ``False`` does not affect the result.
    """
    if isinstance(value, bool):
        raise TypeError("`value` cannot be a bool value")
    perms_value = Permissions(value, **permissions).value

    def decorator(func: T) -> T:
        from .slash_core import SubCommand, SubCommandGroup

        if isinstance(func, InvokableApplicationCommand):
            if isinstance(func, (SubCommand, SubCommandGroup)):
                raise TypeError(
                    "Cannot set `default_member_permissions` on subcommands or subcommand groups"
                )
            if func.body._default_member_permissions is not None:
                raise ValueError(
                    "Cannot set `default_member_permissions` in both parameter and decorator"
                )
            func.body._default_member_permissions = perms_value
        else:
            func.__default_member_permissions__ = perms_value  # type: ignore
        return func

    return decorator


def install_types(*, guild: bool = False, user: bool = False) -> Callable[[T], T]:
    """A decorator that sets the installation types where the
    application command is available.

    See also the ``install_types`` parameter for application command decorators.

    .. note::
        This does not work with slash subcommands/groups.

    .. versionadded:: 2.10

    Parameters
    ----------
    **params: bool
        The installation types; see :class:`.ApplicationInstallTypes`.
        Setting a parameter to ``False`` does not affect the result.
    """

    def decorator(func: T) -> T:
        from .slash_core import SubCommand, SubCommandGroup

        install_types = ApplicationInstallTypes(guild=guild, user=user)
        if isinstance(func, InvokableApplicationCommand):
            if isinstance(func, (SubCommand, SubCommandGroup)):
                raise TypeError("Cannot set `install_types` on subcommands or subcommand groups")
            # special case - don't overwrite if `_guild_only` was set, since that takes priority
            if not func._guild_only:
                if func.body.install_types is not None:
                    raise ValueError("Cannot set `install_types` in both parameter and decorator")
                func.body.install_types = install_types
        else:
            func.__install_types__ = install_types  # type: ignore
        return func

    return decorator


def contexts(
    *, guild: bool = False, bot_dm: bool = False, private_channel: bool = False
) -> Callable[[T], T]:
    """A decorator that sets the interaction contexts where the application command can be used.

    See also the ``contexts`` parameter for application command decorators.

    .. note::
        This does not work with slash subcommands/groups.

    .. versionadded:: 2.10

    Parameters
    ----------
    **params: bool
        The interaction contexts; see :class:`.InteractionContextTypes`.
        Setting a parameter to ``False`` does not affect the result.
    """

    def decorator(func: T) -> T:
        from .slash_core import SubCommand, SubCommandGroup

        contexts = InteractionContextTypes(
            guild=guild, bot_dm=bot_dm, private_channel=private_channel
        )
        if isinstance(func, InvokableApplicationCommand):
            if isinstance(func, (SubCommand, SubCommandGroup)):
                raise TypeError("Cannot set `contexts` on subcommands or subcommand groups")
            # special case - don't overwrite if `_guild_only` was set, since that takes priority
            if not func._guild_only:
                if func.body._dm_permission is not None:
                    raise ValueError(
                        "Cannot use both `dm_permission` and `contexts` at the same time"
                    )
                if func.body.contexts is not None:
                    raise ValueError("Cannot set `contexts` in both parameter and decorator")
                func.body.contexts = contexts
        else:
            func.__contexts__ = contexts  # type: ignore
        return func

    return decorator



# SPDX-License-Identifier: MIT
from typing import Any

import pytest

import disnake
from disnake import Event
from disnake.ext import commands


# n.b. the specific choice of events used in this file is irrelevant
@pytest.fixture
def client():
    return disnake.Client()


@pytest.fixture
def bot():
    return commands.Bot(
        command_prefix=commands.when_mentioned,
        command_sync_flags=commands.CommandSyncFlags.none(),
    )


@pytest.fixture(params=["client", "bot"])
def client_or_bot(request):
    return request.getfixturevalue(request.param)


# @Client.event


def test_event(client_or_bot: disnake.Client) -> None:
    assert not hasattr(client_or_bot, "on_message_edit")

    @client_or_bot.event
    async def on_message_edit(self, *args: Any) -> None:
        ...

    assert client_or_bot.on_message_edit is on_message_edit  # type: ignore


# Client.wait_for


@pytest.mark.parametrize("event", ["thread_create", Event.thread_create])
def test_wait_for(client_or_bot: disnake.Client, event) -> None:
    coro = client_or_bot.wait_for(event)
    assert len(client_or_bot._listeners["thread_create"]) == 1
    coro.close()  # close coroutine to avoid warning


# Client.add_listener / Client.remove_listener


@pytest.mark.parametrize("event", ["on_guild_remove", Event.guild_remove])
def test_addremove_listener(client_or_bot: disnake.Client, event) -> None:
    async def callback(self, *args: Any) -> None:
        ...

    client_or_bot.add_listener(callback, event)
    assert len(client_or_bot.extra_events["on_guild_remove"]) == 1
    client_or_bot.remove_listener(callback, event)
    assert len(client_or_bot.extra_events["on_guild_remove"]) == 0


def test_addremove_listener__implicit(client_or_bot: disnake.Client) -> None:
    async def on_guild_remove(self, *args: Any) -> None:
        ...

    client_or_bot.add_listener(on_guild_remove)
    assert len(client_or_bot.extra_events["on_guild_remove"]) == 1
    client_or_bot.remove_listener(on_guild_remove)
    assert len(client_or_bot.extra_events["on_guild_remove"]) == 0


# @Client.listen


@pytest.mark.parametrize("event", ["on_guild_role_create", Event.guild_role_create])
def test_listen(client_or_bot: disnake.Client, event) -> None:
    @client_or_bot.listen(event)
    async def callback(self, *args: Any) -> None:
        ...

    assert len(client_or_bot.extra_events["on_guild_role_create"]) == 1


def test_listen__implicit(client_or_bot: disnake.Client) -> None:
    @client_or_bot.listen()
    async def on_guild_role_create(self, *args: Any) -> None:
        ...

    assert len(client_or_bot.extra_events["on_guild_role_create"]) == 1


# @commands.Cog.listener
@pytest.mark.parametrize("event", ["on_automod_rule_update", Event.automod_rule_update])
def test_listener(bot: commands.Bot, event) -> None:
    class Cog(commands.Cog):
        @commands.Cog.listener(event)
        async def callback(self, *args: Any) -> None:
            ...

    bot.add_cog(Cog())
    assert len(bot.extra_events["on_automod_rule_update"]) == 1


def test_listener__implicit(bot: commands.Bot) -> None:
    class Cog(commands.Cog):
        @commands.Cog.listener()
        async def on_automod_rule_update(self, *args: Any) -> None:
            ...

    bot.add_cog(Cog())
    assert len(bot.extra_events["on_automod_rule_update"]) == 1



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Dict, List, Literal, Optional, TypedDict

from typing_extensions import NotRequired

from .snowflake import Snowflake
from .team import Team
from .user import User

# (also called "installation context", which seems more accurate)
ApplicationIntegrationType = Literal[0, 1]  # GUILD_INSTALL, USER_INSTALL


class BaseAppInfo(TypedDict):
    id: Snowflake
    name: str
    icon: Optional[str]
    description: str
    terms_of_service_url: NotRequired[str]
    privacy_policy_url: NotRequired[str]
    summary: str
    verify_key: str
    hook: NotRequired[bool]
    max_participants: NotRequired[int]


class InstallParams(TypedDict):
    scopes: List[str]
    permissions: str


class ApplicationIntegrationTypeConfiguration(TypedDict, total=False):
    oauth2_install_params: InstallParams


class AppInfo(BaseAppInfo):
    rpc_origins: List[str]
    bot_public: bool
    bot_require_code_grant: bool
    owner: User
    team: NotRequired[Team]
    guild_id: NotRequired[Snowflake]
    primary_sku_id: NotRequired[Snowflake]
    slug: NotRequired[str]
    tags: NotRequired[List[str]]
    install_params: NotRequired[InstallParams]
    custom_install_url: NotRequired[str]
    role_connections_verification_url: NotRequired[str]
    approximate_guild_count: NotRequired[int]
    approximate_user_install_count: NotRequired[int]
    # values in this dict generally shouldn't be null, but they can be empty dicts
    integration_types_config: NotRequired[Dict[str, ApplicationIntegrationTypeConfiguration]]


class PartialAppInfo(BaseAppInfo, total=False):
    rpc_origins: List[str]
    cover_image: str
    flags: int


# see https://discord.com/developers/docs/topics/gateway-events#ready-ready-event-fields
class PartialGatewayAppInfo(TypedDict):
    id: Snowflake
    flags: int



# SPDX-License-Identifier: MIT

from __future__ import annotations

import colorsys
import random
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

if TYPE_CHECKING:
    from typing_extensions import Self


__all__ = (
    "Colour",
    "Color",
)


class Colour:
    """Represents a Discord role colour. This class is similar
    to a (red, green, blue) :class:`tuple`.

    There is an alias for this called Color.

    .. collapse:: operations

        .. describe:: x == y

             Checks if two colours are equal.

        .. describe:: x != y

             Checks if two colours are not equal.

        .. describe:: hash(x)

             Return the colour's hash.

        .. describe:: str(x)

             Returns the hex format for the colour.

        .. describe:: int(x)

             Returns the raw colour value.

    Attributes
    ----------
    value: :class:`int`
        The raw integer colour value.
    """

    __slots__ = ("value",)

    def __init__(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError(f"Expected int parameter, received {value.__class__.__name__} instead.")

        self.value: int = value

    def _get_byte(self, byte: int) -> int:
        return (self.value >> (8 * byte)) & 0xFF

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Colour) and self.value == other.value

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"#{self.value:0>6x}"

    def __int__(self) -> int:
        return self.value

    def __repr__(self) -> str:
        return f"<Colour value={self.value}>"

    def __hash__(self) -> int:
        return hash(self.value)

    @property
    def r(self) -> int:
        """:class:`int`: Returns the red component of the colour."""
        return self._get_byte(2)

    @property
    def g(self) -> int:
        """:class:`int`: Returns the green component of the colour."""
        return self._get_byte(1)

    @property
    def b(self) -> int:
        """:class:`int`: Returns the blue component of the colour."""
        return self._get_byte(0)

    def to_rgb(self) -> Tuple[int, int, int]:
        """Tuple[:class:`int`, :class:`int`, :class:`int`]: Returns an (r, g, b) tuple representing the colour."""
        return (self.r, self.g, self.b)

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int) -> Self:
        """Constructs a :class:`Colour` from an RGB tuple."""
        return cls((r << 16) + (g << 8) + b)

    @classmethod
    def from_hsv(cls, h: float, s: float, v: float) -> Self:
        """Constructs a :class:`Colour` from an HSV tuple."""
        rgb = colorsys.hsv_to_rgb(h, s, v)
        return cls.from_rgb(*(int(x * 255) for x in rgb))

    @classmethod
    def default(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0``."""
        return cls(0)

    @classmethod
    def random(cls, *, seed: Optional[Union[int, str, float, bytes, bytearray]] = None) -> Self:
        """A factory method that returns a :class:`Colour` with a random hue.

        .. note::

            The random algorithm works by choosing a colour with a random hue but
            with maxed out saturation and value.

        .. versionadded:: 1.6

        Parameters
        ----------
        seed: Optional[Union[:class:`int`, :class:`str`, :class:`float`, :class:`bytes`, :class:`bytearray`]]
            The seed to initialize the RNG with. If ``None`` is passed the default RNG is used.

            .. versionadded:: 1.7
        """
        rand = random if seed is None else random.Random(seed)
        return cls.from_hsv(rand.random(), 1, 1)

    @classmethod
    def teal(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x1abc9c``."""
        return cls(0x1ABC9C)

    @classmethod
    def dark_teal(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x11806a``."""
        return cls(0x11806A)

    @classmethod
    def brand_green(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x57F287``.

        .. versionadded:: 2.0
        """
        return cls(0x57F287)

    @classmethod
    def green(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x2ecc71``."""
        return cls(0x2ECC71)

    @classmethod
    def dark_green(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x1f8b4c``."""
        return cls(0x1F8B4C)

    @classmethod
    def blue(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x3498db``."""
        return cls(0x3498DB)

    @classmethod
    def dark_blue(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x206694``."""
        return cls(0x206694)

    @classmethod
    def purple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x9b59b6``."""
        return cls(0x9B59B6)

    @classmethod
    def dark_purple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x71368a``."""
        return cls(0x71368A)

    @classmethod
    def magenta(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xe91e63``."""
        return cls(0xE91E63)

    @classmethod
    def dark_magenta(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xad1457``."""
        return cls(0xAD1457)

    @classmethod
    def gold(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xf1c40f``."""
        return cls(0xF1C40F)

    @classmethod
    def dark_gold(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xc27c0e``."""
        return cls(0xC27C0E)

    @classmethod
    def orange(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xe67e22``."""
        return cls(0xE67E22)

    @classmethod
    def dark_orange(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xa84300``."""
        return cls(0xA84300)

    @classmethod
    def brand_red(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xED4245``.

        .. versionadded:: 2.0
        """
        return cls(0xED4245)

    @classmethod
    def red(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xe74c3c``."""
        return cls(0xE74C3C)

    @classmethod
    def dark_red(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x992d22``."""
        return cls(0x992D22)

    @classmethod
    def lighter_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x95a5a6``."""
        return cls(0x95A5A6)

    lighter_gray = lighter_grey

    @classmethod
    def dark_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x607d8b``."""
        return cls(0x607D8B)

    dark_gray = dark_grey

    @classmethod
    def light_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x979c9f``."""
        return cls(0x979C9F)

    light_gray = light_grey

    @classmethod
    def darker_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x546e7a``."""
        return cls(0x546E7A)

    darker_gray = darker_grey

    @classmethod
    def og_blurple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x7289da``."""
        return cls(0x7289DA)

    old_blurple = og_blurple

    @classmethod
    def blurple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x5865F2``."""
        return cls(0x5865F2)

    @classmethod
    def greyple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x99aab5``."""
        return cls(0x99AAB5)

    @classmethod
    def dark_theme(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x313338``.
        This will appear transparent on Discord's dark theme.

        .. versionadded:: 1.5
        """
        return cls(0x313338)

    @classmethod
    def fuchsia(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xEB459E``.

        .. versionadded:: 2.0
        """
        return cls(0xEB459E)

    @classmethod
    def yellow(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xFEE75C``.

        .. versionadded:: 2.0
        """
        return cls(0xFEE75C)

    @classmethod
    def light_embed(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xF2F3F5``.
        This matches the embed background colour on Discord's light theme.

        .. versionadded:: 2.10
        """
        return cls(0xF2F3F5)

    @classmethod
    def dark_embed(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x2B2D31``.
        This matches the embed background colour on Discord's dark theme.

        .. versionadded:: 2.10
        """
        return cls(0x2B2D31)


Color = Colour



# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

from .enums import ExpireBehaviour, try_enum
from .user import User
from .utils import (
    MISSING,
    _get_as_snowflake,
    deprecated,
    parse_time,
    snowflake_time,
    warn_deprecated,
)

__all__ = (
    "IntegrationAccount",
    "IntegrationApplication",
    "PartialIntegration",
    "Integration",
    "StreamIntegration",
    "BotIntegration",
)

if TYPE_CHECKING:
    from .guild import Guild
    from .role import Role
    from .types.integration import (
        BotIntegration as BotIntegrationPayload,
        Integration as IntegrationPayload,
        IntegrationAccount as IntegrationAccountPayload,
        IntegrationApplication as IntegrationApplicationPayload,
        IntegrationType,
        PartialIntegration as PartialIntegrationPayload,
        StreamIntegration as StreamIntegrationPayload,
    )


class IntegrationAccount:
    """Represents an integration account.

    .. versionadded:: 1.4

    Attributes
    ----------
    id: :class:`str`
        The account ID.
    name: :class:`str`
        The account name.
    """

    __slots__ = ("id", "name")

    def __init__(self, data: IntegrationAccountPayload) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]

    def __repr__(self) -> str:
        return f"<IntegrationAccount id={self.id} name={self.name!r}>"


class PartialIntegration:
    """Represents a partial guild integration.

    .. versionadded:: 2.6

    Attributes
    ----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type (i.e. ``twitch``).
    account: :class:`IntegrationAccount`
        The account linked to this integration.
    application_id: Optional[:class:`int`]
        The ID of the application tied to this integration.
    """

    __slots__ = (
        "guild",
        "id",
        "name",
        "type",
        "account",
        "application_id",
    )

    def __init__(self, *, data: PartialIntegrationPayload, guild: Guild) -> None:
        self.guild = guild
        self._from_data(data)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name!r}>"

    def _from_data(self, data: PartialIntegrationPayload) -> None:
        self.id: int = int(data["id"])
        self.type: IntegrationType = data["type"]
        self.name: str = data["name"]
        self.account: IntegrationAccount = IntegrationAccount(data["account"])
        self.application_id: Optional[int] = _get_as_snowflake(data, "application_id")

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the integration's
        (*not* the associated application's) creation time in UTC.

        .. versionadded:: 2.10
        """
        return snowflake_time(self.id)


class Integration(PartialIntegration):
    """Represents a guild integration.

    .. versionadded:: 1.4

    Attributes
    ----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type (i.e. Twitch).
    enabled: :class:`bool`
        Whether the integration is currently enabled.
    account: :class:`IntegrationAccount`
        The account linked to this integration.
    user: Optional[:class:`User`]
        The user that added this integration.
    """

    __slots__ = (
        "_state",
        "user",
        "enabled",
    )

    def _from_data(self, data: IntegrationPayload) -> None:
        super()._from_data(data)
        self._state = self.guild._state

        user = data.get("user")
        self.user = User(state=self._state, data=user) if user else None
        self.enabled: bool = data["enabled"]

    @deprecated("Guild.leave")
    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        .. deprecated:: 2.5
            Can only be used on the application's own integration and is therefore
            equivalent to leaving the guild.

        Deletes the integration.

        You must have :attr:`~Permissions.manage_guild` permission to
        use this.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason the integration was deleted. Shows up on the audit log.

            .. versionadded:: 2.0

        Raises
        ------
        Forbidden
            You do not have permission to delete the integration.
        HTTPException
            Deleting the integration failed.
        """
        await self._state.http.delete_integration(self.guild.id, self.id, reason=reason)


class StreamIntegration(Integration):
    """Represents a stream integration for Twitch or YouTube.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type (i.e. Twitch).
    enabled: :class:`bool`
        Whether the integration is currently enabled.
    syncing: :class:`bool`
        Whether the integration is currently syncing.
    enable_emoticons: Optional[:class:`bool`]
        Whether emoticons should be synced for this integration (currently twitch only).
    expire_behaviour: :class:`ExpireBehaviour`
        The behaviour of expiring subscribers. Aliased to ``expire_behavior`` as well.
    expire_grace_period: :class:`int`
        The grace period (in days) for expiring subscribers.
    user: :class:`User`
        The user for the integration.
    account: :class:`IntegrationAccount`
        The integration account information.
    synced_at: :class:`datetime.datetime`
        An aware UTC datetime representing when the integration was last synced.
    """

    __slots__ = (
        "revoked",
        "expire_behaviour",
        "expire_grace_period",
        "synced_at",
        "_role_id",
        "syncing",
        "enable_emoticons",
        "subscriber_count",
    )

    def _from_data(self, data: StreamIntegrationPayload) -> None:
        super()._from_data(data)
        self.revoked: bool = data["revoked"]
        self.expire_behaviour: ExpireBehaviour = try_enum(ExpireBehaviour, data["expire_behavior"])
        self.expire_grace_period: int = data["expire_grace_period"]
        self.synced_at: datetime.datetime = parse_time(data["synced_at"])
        self._role_id: Optional[int] = _get_as_snowflake(data, "role_id")
        self.syncing: bool = data["syncing"]
        self.enable_emoticons: bool = data["enable_emoticons"]
        self.subscriber_count: int = data["subscriber_count"]

    @property
    def expire_behavior(self) -> ExpireBehaviour:
        """:class:`ExpireBehaviour`: An alias for :attr:`expire_behaviour`."""
        return self.expire_behaviour

    @property
    def role(self) -> Optional[Role]:
        """Optional[:class:`Role`] The role which the integration uses for subscribers."""
        return self.guild.get_role(self._role_id)  # type: ignore

    @deprecated()
    async def edit(
        self,
        *,
        expire_behaviour: ExpireBehaviour = MISSING,
        expire_grace_period: int = MISSING,
        enable_emoticons: bool = MISSING,
    ) -> None:
        """|coro|

        .. deprecated:: 2.5
            No longer supported, bots cannot use this endpoint anymore.

        Edits the integration.

        You must have :attr:`~Permissions.manage_guild` permission to
        use this.

        .. versionchanged:: 2.6
            Raises :exc:`TypeError` instead of ``InvalidArgument``.

        Parameters
        ----------
        expire_behaviour: :class:`ExpireBehaviour`
            The behaviour when an integration subscription lapses. Aliased to ``expire_behavior`` as well.
        expire_grace_period: :class:`int`
            The period (in days) where the integration will ignore lapsed subscriptions.
        enable_emoticons: :class:`bool`
            Where emoticons should be synced for this integration (currently twitch only).

        Raises
        ------
        Forbidden
            You do not have permission to edit the integration.
        HTTPException
            Editing the guild failed.
        TypeError
            ``expire_behaviour`` did not receive a :class:`ExpireBehaviour`.
        """
        payload: Dict[str, Any] = {}
        if expire_behaviour is not MISSING:
            if not isinstance(expire_behaviour, ExpireBehaviour):
                raise TypeError("expire_behaviour field must be of type ExpireBehaviour")

            payload["expire_behavior"] = expire_behaviour.value

        if expire_grace_period is not MISSING:
            payload["expire_grace_period"] = expire_grace_period

        if enable_emoticons is not MISSING:
            payload["enable_emoticons"] = enable_emoticons

        # This endpoint is undocumented.
        # Unsure if it returns the data or not as a result
        await self._state.http.edit_integration(self.guild.id, self.id, **payload)

    @deprecated()
    async def sync(self) -> None:
        """|coro|

        .. deprecated:: 2.5
            No longer supported, bots cannot use this endpoint anymore.

        Syncs the integration.

        You must have :attr:`~Permissions.manage_guild` permission to
        use this.

        Raises
        ------
        Forbidden
            You do not have permission to sync the integration.
        HTTPException
            Syncing the integration failed.
        """
        await self._state.http.sync_integration(self.guild.id, self.id)
        self.synced_at = datetime.datetime.now(datetime.timezone.utc)


class IntegrationApplication:
    """Represents an application for a bot integration.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The application's ID.
    name: :class:`str`
        The application's name.
    icon: Optional[:class:`str`]
        The application's icon hash.
    description: :class:`str`
        The application's description. Can be an empty string.
    user: Optional[:class:`User`]
        The bot user associated with this application.
    """

    __slots__ = (
        "id",
        "name",
        "icon",
        "description",
        "_summary",
        "user",
    )

    def __init__(self, *, data: IntegrationApplicationPayload, state) -> None:
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self.icon: Optional[str] = data["icon"]
        self.description: str = data["description"]
        self._summary: str = data.get("summary", "")
        user = data.get("bot")
        self.user: Optional[User] = User(state=state, data=user) if user else None

    @property
    def summary(self) -> str:
        """:class:`str`: The application's summary. Can be an empty string.

        .. deprecated:: 2.5

            This field is deprecated by discord and is now always blank. Consider using :attr:`.description` instead.
        """
        warn_deprecated(
            "summary is deprecated and will be removed in a future version. Consider using description instead.",
            stacklevel=2,
        )
        return self._summary


class BotIntegration(Integration):
    """Represents a bot integration on Discord.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The integration ID.
    name: :class:`str`
        The integration name.
    guild: :class:`Guild`
        The guild of the integration.
    type: :class:`str`
        The integration type (i.e. Twitch).
    enabled: :class:`bool`
        Whether the integration is currently enabled.
    user: :class:`User`
        The user that added this integration.
    account: :class:`IntegrationAccount`
        The integration account information.
    application: :class:`IntegrationApplication`
        The application tied to this integration.
    scopes: List[:class:`str`]
        The OAuth2 scopes the application has been authorized for.

        .. versionadded:: 2.6
    """

    __slots__ = ("application", "scopes")

    def _from_data(self, data: BotIntegrationPayload) -> None:
        super()._from_data(data)
        self.application: IntegrationApplication = IntegrationApplication(
            data=data["application"], state=self._state
        )
        self.scopes: List[str] = data.get("scopes") or []

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id}"
            f" name={self.name!r} scopes={self.scopes!r}>"
        )


def _integration_factory(value: str) -> Tuple[Type[Integration], str]:
    if value == "discord":
        return BotIntegration, value
    elif value in ("twitch", "youtube"):
        return StreamIntegration, value
    else:
        return Integration, value



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Type, Union

from disnake.errors import ClientException, DiscordException
from disnake.utils import humanize_list

if TYPE_CHECKING:
    from inspect import Parameter

    from disnake.abc import GuildChannel
    from disnake.threads import Thread
    from disnake.types.snowflake import Snowflake, SnowflakeList

    from .context import AnyContext
    from .cooldowns import BucketType, Cooldown
    from .flag_converter import Flag


__all__ = (
    "CommandError",
    "MissingRequiredArgument",
    "BadArgument",
    "PrivateMessageOnly",
    "NoPrivateMessage",
    "CheckFailure",
    "CheckAnyFailure",
    "CommandNotFound",
    "DisabledCommand",
    "CommandInvokeError",
    "TooManyArguments",
    "UserInputError",
    "CommandOnCooldown",
    "MaxConcurrencyReached",
    "NotOwner",
    "MessageNotFound",
    "ObjectNotFound",
    "MemberNotFound",
    "GuildNotFound",
    "UserNotFound",
    "ChannelNotFound",
    "ThreadNotFound",
    "ChannelNotReadable",
    "BadColourArgument",
    "BadColorArgument",
    "RoleNotFound",
    "BadInviteArgument",
    "EmojiNotFound",
    "GuildStickerNotFound",
    "GuildSoundboardSoundNotFound",
    "GuildScheduledEventNotFound",
    "PartialEmojiConversionFailure",
    "BadBoolArgument",
    "LargeIntConversionFailure",
    "MissingRole",
    "BotMissingRole",
    "MissingAnyRole",
    "BotMissingAnyRole",
    "MissingPermissions",
    "BotMissingPermissions",
    "NSFWChannelRequired",
    "ConversionError",
    "BadUnionArgument",
    "BadLiteralArgument",
    "ArgumentParsingError",
    "UnexpectedQuoteError",
    "InvalidEndOfQuotedStringError",
    "ExpectedClosingQuoteError",
    "ExtensionError",
    "ExtensionAlreadyLoaded",
    "ExtensionNotLoaded",
    "NoEntryPointError",
    "ExtensionFailed",
    "ExtensionNotFound",
    "CommandRegistrationError",
    "FlagError",
    "BadFlagArgument",
    "MissingFlagArgument",
    "TooManyFlags",
    "MissingRequiredFlag",
)


class CommandError(DiscordException):
    """The base exception type for all command related errors.

    This inherits from :exc:`disnake.DiscordException`.

    This exception and exceptions inherited from it are handled
    in a special way as they are caught and passed into a special event
    from :class:`.Bot`\\, :func:`.on_command_error`.
    """

    def __init__(self, message: Optional[str] = None, *args: Any) -> None:
        if message is not None:
            # clean-up @everyone and @here mentions
            m = message.replace("@everyone", "@\u200beveryone").replace("@here", "@\u200bhere")
            super().__init__(m, *args)
        else:
            super().__init__(*args)


class ConversionError(CommandError):
    """Exception raised when a Converter class raises non-CommandError.

    This inherits from :exc:`CommandError`.

    Attributes
    ----------
    converter: :class:`.Converter`
        The converter that failed.
    original: :exc:`Exception`
        The original exception that was raised. You can also get this via
        the ``__cause__`` attribute.
    """

    def __init__(self, converter: Any, original: Exception) -> None:
        self.converter: Any = converter
        self.original: Exception = original


class UserInputError(CommandError):
    """The base exception type for errors that involve errors
    regarding user input.

    This inherits from :exc:`CommandError`.
    """

    pass


class CommandNotFound(CommandError):
    """Exception raised when a command is attempted to be invoked
    but no command under that name is found.

    This is not raised for invalid subcommands, rather just the
    initial main command that is attempted to be invoked.

    This inherits from :exc:`CommandError`.
    """

    pass


class MissingRequiredArgument(UserInputError):
    """Exception raised when parsing a command and a parameter
    that is required is not encountered.

    This inherits from :exc:`UserInputError`

    Attributes
    ----------
    param: :class:`inspect.Parameter`
        The argument that is missing.
    """

    def __init__(self, param: Parameter) -> None:
        self.param: Parameter = param
        super().__init__(f"{param.name} is a required argument that is missing.")


class TooManyArguments(UserInputError):
    """Exception raised when the command was passed too many arguments and its
    :attr:`.Command.ignore_extra` attribute was not set to ``True``.

    This inherits from :exc:`UserInputError`
    """

    pass


class BadArgument(UserInputError):
    """Exception raised when a parsing or conversion failure is encountered
    on an argument to pass into a command.

    This inherits from :exc:`UserInputError`
    """

    pass


class CheckFailure(CommandError):
    """Exception raised when the predicates in :attr:`.Command.checks` or
    :attr:`.InvokableApplicationCommand.checks` have failed.

    This inherits from :exc:`CommandError`
    """

    pass


class CheckAnyFailure(CheckFailure):
    """Exception raised when all predicates in :func:`check_any` or :func:`app_check_any` fail.

    This inherits from :exc:`CheckFailure`.

    .. versionadded:: 1.3

    Attributes
    ----------
    errors: List[:class:`CheckFailure`]
        A list of errors that were caught during execution.
    checks: List[Callable[[Union[:class:`Context`, :class:`disnake.ApplicationCommandInteraction`]], :class:`bool`]]
        A list of check predicates that failed.
    """

    def __init__(
        self, checks: List[CheckFailure], errors: List[Callable[[AnyContext], bool]]
    ) -> None:
        self.checks: List[CheckFailure] = checks
        self.errors: List[Callable[[AnyContext], bool]] = errors
        super().__init__("You do not have permission to run this command.")


class PrivateMessageOnly(CheckFailure):
    """Exception raised when an operation does not work outside of private
    message contexts.

    This inherits from :exc:`CheckFailure`
    """

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "This command can only be used in private messages.")


class NoPrivateMessage(CheckFailure):
    """Exception raised when an operation does not work in private message
    contexts.

    This inherits from :exc:`CheckFailure`
    """

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "This command cannot be used in private messages.")


class NotOwner(CheckFailure):
    """Exception raised when the message author is not the owner of the bot.

    This inherits from :exc:`CheckFailure`
    """

    pass


class ObjectNotFound(BadArgument):
    """Exception raised when the argument provided did not match the format
    of an ID or a mention.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 2.0

    Attributes
    ----------
    argument: :class:`str`
        The argument supplied by the caller that was not matched
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f"{argument!r} does not follow a valid ID or mention format.")


class MemberNotFound(BadArgument):
    """Exception raised when the member provided was not found in the bot's
    cache.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The member supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Member "{argument}" not found.')


class GuildNotFound(BadArgument):
    """Exception raised when the guild provided was not found in the bot's cache.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.7

    Attributes
    ----------
    argument: :class:`str`
        The guild supplied by the called that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Guild "{argument}" not found.')


class UserNotFound(BadArgument):
    """Exception raised when the user provided was not found in the bot's
    cache.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The user supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'User "{argument}" not found.')


class MessageNotFound(BadArgument):
    """Exception raised when the message provided was not found in the channel.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The message supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Message "{argument}" not found.')


class ChannelNotReadable(BadArgument):
    """Exception raised when the bot does not have permission to read messages
    in the channel.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: Union[:class:`.abc.GuildChannel`, :class:`.Thread`]
        The channel supplied by the caller that was not readable
    """

    def __init__(self, argument: Union[GuildChannel, Thread]) -> None:
        self.argument: Union[GuildChannel, Thread] = argument
        super().__init__(f"Can't read messages in {argument.mention}.")


class ChannelNotFound(BadArgument):
    """Exception raised when the bot can not find the channel.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The channel supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Channel "{argument}" not found.')


class ThreadNotFound(BadArgument):
    """Exception raised when the bot can not find the thread.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 2.0

    Attributes
    ----------
    argument: :class:`str`
        The thread supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Thread "{argument}" not found.')


class BadColourArgument(BadArgument):
    """Exception raised when the colour is not valid.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The colour supplied by the caller that was not valid
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Colour "{argument}" is invalid.')


BadColorArgument = BadColourArgument


class RoleNotFound(BadArgument):
    """Exception raised when the bot can not find the role.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The role supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Role "{argument}" not found.')


class BadInviteArgument(BadArgument):
    """Exception raised when the invite is invalid or expired.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Invite "{argument}" is invalid or expired.')


class EmojiNotFound(BadArgument):
    """Exception raised when the bot can not find the emoji.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The emoji supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Emoji "{argument}" not found.')


class PartialEmojiConversionFailure(BadArgument):
    """Exception raised when the emoji provided does not match the correct
    format.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The emoji supplied by the caller that did not match the regex
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Couldn\'t convert "{argument}" to PartialEmoji.')


class GuildStickerNotFound(BadArgument):
    """Exception raised when the bot can not find the sticker.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 2.0

    Attributes
    ----------
    argument: :class:`str`
        The sticker supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Sticker "{argument}" not found.')


class GuildSoundboardSoundNotFound(BadArgument):
    """Exception raised when the bot can not find the soundboard sound.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 2.10

    Attributes
    ----------
    argument: :class:`str`
        The soundboard sound supplied by the caller that was not found
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Soundboard sound "{argument}" not found.')


class GuildScheduledEventNotFound(BadArgument):
    """Exception raised when the bot cannot find the scheduled event.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 2.5

    Attributes
    ----------
    argument: :class:`str`
        The scheduled event ID/URL/name supplied by the caller that was not found.
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f'Scheduled event "{argument}" not found.')


class BadBoolArgument(BadArgument):
    """Exception raised when a boolean argument was not convertable.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 1.5

    Attributes
    ----------
    argument: :class:`str`
        The boolean argument supplied by the caller that is not in the predefined list
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f"{argument} is not a recognised boolean option")


class LargeIntConversionFailure(BadArgument):
    """Exception raised when a large integer argument was not able to be converted.

    This inherits from :exc:`BadArgument`

    .. versionadded:: 2.5

    Attributes
    ----------
    argument: :class:`str`
        The argument that could not be converted to an integer.
    """

    def __init__(self, argument: str) -> None:
        self.argument: str = argument
        super().__init__(f"{argument} is not able to be converted to an integer")


class DisabledCommand(CommandError):
    """Exception raised when the command being invoked is disabled.

    This inherits from :exc:`CommandError`
    """

    pass


class CommandInvokeError(CommandError):
    """Exception raised when the command being invoked raised an exception.

    This inherits from :exc:`CommandError`

    Attributes
    ----------
    original: :exc:`Exception`
        The original exception that was raised. You can also get this via
        the ``__cause__`` attribute.
    """

    def __init__(self, e: Exception) -> None:
        self.original: Exception = e
        super().__init__(f"Command raised an exception: {e.__class__.__name__}: {e}")


class CommandOnCooldown(CommandError):
    """Exception raised when the command being invoked is on cooldown.

    This inherits from :exc:`CommandError`

    Attributes
    ----------
    cooldown: :class:`.Cooldown`
        A class with attributes ``rate`` and ``per`` similar to the
        :func:`.cooldown` decorator.
    type: :class:`BucketType`
        The type associated with the cooldown.
    retry_after: :class:`float`
        The amount of seconds to wait before you can retry again.
    """

    def __init__(self, cooldown: Cooldown, retry_after: float, type: BucketType) -> None:
        self.cooldown: Cooldown = cooldown
        self.retry_after: float = retry_after
        self.type: BucketType = type
        super().__init__(f"You are on cooldown. Try again in {retry_after:.2f}s")


class MaxConcurrencyReached(CommandError):
    """Exception raised when the command being invoked has reached its maximum concurrency.

    This inherits from :exc:`CommandError`.

    Attributes
    ----------
    number: :class:`int`
        The maximum number of concurrent invokers allowed.
    per: :class:`.BucketType`
        The bucket type passed to the :func:`.max_concurrency` decorator.
    """

    def __init__(self, number: int, per: BucketType) -> None:
        self.number: int = number
        self.per: BucketType = per
        name = per.name
        suffix = f"per {name}" if per.name != "default" else "globally"
        fmt = f"{number} times {suffix}" if number > 1 else f"{number} time {suffix}"
        super().__init__(
            f"Too many people are using this command. It can only be used {fmt} concurrently."
        )


class MissingRole(CheckFailure):
    """Exception raised when the command invoker lacks a role to run a command.

    This inherits from :exc:`CheckFailure`

    .. versionadded:: 1.1

    Attributes
    ----------
    missing_role: Union[:class:`str`, :class:`int`]
        The required role that is missing.
        This is the parameter passed to :func:`~.commands.has_role`.
    """

    def __init__(self, missing_role: Snowflake) -> None:
        self.missing_role: Snowflake = missing_role
        message = f"Role {missing_role!r} is required to run this command."
        super().__init__(message)


class BotMissingRole(CheckFailure):
    """Exception raised when the bot's member lacks a role to run a command.

    This inherits from :exc:`CheckFailure`

    .. versionadded:: 1.1

    Attributes
    ----------
    missing_role: Union[:class:`str`, :class:`int`]
        The required role that is missing.
        This is the parameter passed to :func:`~.commands.has_role`.
    """

    def __init__(self, missing_role: Snowflake) -> None:
        self.missing_role: Snowflake = missing_role
        message = f"Bot requires the role {missing_role!r} to run this command"
        super().__init__(message)


class MissingAnyRole(CheckFailure):
    """Exception raised when the command invoker lacks any of
    the roles specified to run a command.

    This inherits from :exc:`CheckFailure`

    .. versionadded:: 1.1

    Attributes
    ----------
    missing_roles: List[Union[:class:`str`, :class:`int`]]
        The roles that the invoker is missing.
        These are the parameters passed to :func:`~.commands.has_any_role`.
    """

    def __init__(self, missing_roles: SnowflakeList) -> None:
        self.missing_roles: SnowflakeList = missing_roles

        missing = [f"'{role}'" for role in missing_roles]
        fmt = humanize_list(missing, "or")

        message = f"You are missing at least one of the required roles: {fmt}"
        super().__init__(message)


class BotMissingAnyRole(CheckFailure):
    """Exception raised when the bot's member lacks any of
    the roles specified to run a command.

    This inherits from :exc:`CheckFailure`

    .. versionadded:: 1.1

    Attributes
    ----------
    missing_roles: List[Union[:class:`str`, :class:`int`]]
        The roles that the bot's member is missing.
        These are the parameters passed to :func:`~.commands.has_any_role`.

    """

    def __init__(self, missing_roles: SnowflakeList) -> None:
        self.missing_roles: SnowflakeList = missing_roles

        missing = [f"'{role}'" for role in missing_roles]
        fmt = humanize_list(missing, "or")

        message = f"Bot is missing at least one of the required roles: {fmt}"
        super().__init__(message)


class NSFWChannelRequired(CheckFailure):
    """Exception raised when a channel does not have the required NSFW setting.

    This inherits from :exc:`CheckFailure`.

    .. versionadded:: 1.1

    Parameters
    ----------
    channel: Union[:class:`.abc.GuildChannel`, :class:`.Thread`]
        The channel that does not have NSFW enabled.
    """

    def __init__(self, channel: Union[GuildChannel, Thread]) -> None:
        self.channel: Union[GuildChannel, Thread] = channel
        super().__init__(f"Channel '{channel}' needs to be NSFW for this command to work.")


class MissingPermissions(CheckFailure):
    """Exception raised when the command invoker lacks permissions to run a
    command.

    This inherits from :exc:`CheckFailure`

    Attributes
    ----------
    missing_permissions: List[:class:`str`]
        The required permissions that are missing.
    """

    def __init__(self, missing_permissions: List[str], *args: Any) -> None:
        self.missing_permissions: List[str] = missing_permissions

        missing = [
            perm.replace("_", " ").replace("guild", "server").title()
            for perm in missing_permissions
        ]
        fmt = humanize_list(missing, "and")

        message = f"You are missing {fmt} permission(s) to run this command."
        super().__init__(message, *args)


class BotMissingPermissions(CheckFailure):
    """Exception raised when the bot's member lacks permissions to run a
    command.

    This inherits from :exc:`CheckFailure`

    Attributes
    ----------
    missing_permissions: List[:class:`str`]
        The required permissions that are missing.
    """

    def __init__(self, missing_permissions: List[str], *args: Any) -> None:
        self.missing_permissions: List[str] = missing_permissions

        missing = [
            perm.replace("_", " ").replace("guild", "server").title()
            for perm in missing_permissions
        ]
        fmt = humanize_list(missing, "and")

        message = f"Bot requires {fmt} permission(s) to run this command."
        super().__init__(message, *args)


class BadUnionArgument(UserInputError):
    """Exception raised when a :data:`typing.Union` converter fails for all
    its associated types.

    This inherits from :exc:`UserInputError`

    Attributes
    ----------
    param: :class:`inspect.Parameter`
        The parameter that failed being converted.
    converters: Tuple[Type, ``...``]
        A tuple of converters attempted in conversion, in order of failure.
    errors: List[:class:`CommandError`]
        A list of errors that were caught from failing the conversion.
    """

    def __init__(
        self, param: Parameter, converters: Tuple[Type, ...], errors: List[CommandError]
    ) -> None:
        self.param: Parameter = param
        self.converters: Tuple[Type, ...] = converters
        self.errors: List[CommandError] = errors

        def _get_name(x):
            try:
                return x.__name__
            except AttributeError:
                if hasattr(x, "__origin__"):
                    return repr(x)
                return x.__class__.__name__

        to_string = [_get_name(x) for x in converters]
        fmt = humanize_list(to_string, "or")

        super().__init__(f'Could not convert "{param.name}" into {fmt}.')


class BadLiteralArgument(UserInputError):
    """Exception raised when a :data:`typing.Literal` converter fails for all
    its associated values.

    This inherits from :exc:`UserInputError`

    .. versionadded:: 2.0

    Attributes
    ----------
    param: :class:`inspect.Parameter`
        The parameter that failed being converted.
    literals: Tuple[Any, ``...``]
        A tuple of values compared against in conversion, in order of failure.
    errors: List[:class:`CommandError`]
        A list of errors that were caught from failing the conversion.
    """

    def __init__(
        self, param: Parameter, literals: Tuple[Any, ...], errors: List[CommandError]
    ) -> None:
        self.param: Parameter = param
        self.literals: Tuple[Any, ...] = literals
        self.errors: List[CommandError] = errors

        to_string = [repr(literal) for literal in literals]
        fmt = humanize_list(to_string, "or")

        super().__init__(f'Could not convert "{param.name}" into the literal {fmt}.')


class ArgumentParsingError(UserInputError):
    """An exception raised when the parser fails to parse a user's input.

    This inherits from :exc:`UserInputError`.

    There are child classes that implement more granular parsing errors for
    i18n purposes.
    """

    pass


class UnexpectedQuoteError(ArgumentParsingError):
    """An exception raised when the parser encounters a quote mark inside a non-quoted string.

    This inherits from :exc:`ArgumentParsingError`.

    Attributes
    ----------
    quote: :class:`str`
        The quote mark that was found inside the non-quoted string.
    """

    def __init__(self, quote: str) -> None:
        self.quote: str = quote
        super().__init__(f"Unexpected quote mark, {quote!r}, in non-quoted string")


class InvalidEndOfQuotedStringError(ArgumentParsingError):
    """An exception raised when a space is expected after the closing quote in a string
    but a different character is found.

    This inherits from :exc:`ArgumentParsingError`.

    Attributes
    ----------
    char: :class:`str`
        The character found instead of the expected string.
    """

    def __init__(self, char: str) -> None:
        self.char: str = char
        super().__init__(f"Expected space after closing quotation but received {char!r}")


class ExpectedClosingQuoteError(ArgumentParsingError):
    """An exception raised when a quote character is expected but not found.

    This inherits from :exc:`ArgumentParsingError`.

    Attributes
    ----------
    close_quote: :class:`str`
        The quote character expected.
    """

    def __init__(self, close_quote: str) -> None:
        self.close_quote: str = close_quote
        super().__init__(f"Expected closing {close_quote}.")


class ExtensionError(DiscordException):
    """Base exception for extension related errors.

    This inherits from :exc:`~disnake.DiscordException`.

    Attributes
    ----------
    name: :class:`str`
        The extension that had an error.
    """

    def __init__(self, message: Optional[str] = None, *args: Any, name: str) -> None:
        self.name: str = name
        message = message or f"Extension {name!r} had an error."
        # clean-up @everyone and @here mentions
        m = message.replace("@everyone", "@\u200beveryone").replace("@here", "@\u200bhere")
        super().__init__(m, *args)


class ExtensionAlreadyLoaded(ExtensionError):
    """An exception raised when an extension has already been loaded.

    This inherits from :exc:`ExtensionError`
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Extension {name!r} is already loaded.", name=name)


class ExtensionNotLoaded(ExtensionError):
    """An exception raised when an extension was not loaded.

    This inherits from :exc:`ExtensionError`
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Extension {name!r} has not been loaded.", name=name)


class NoEntryPointError(ExtensionError):
    """An exception raised when an extension does not have a ``setup`` entry point function.

    This inherits from :exc:`ExtensionError`
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Extension {name!r} has no 'setup' function.", name=name)


class ExtensionFailed(ExtensionError):
    """An exception raised when an extension failed to load during execution of the module or ``setup`` entry point.

    This inherits from :exc:`ExtensionError`

    Attributes
    ----------
    name: :class:`str`
        The extension that had the error.
    original: :exc:`Exception`
        The original exception that was raised. You can also get this via
        the ``__cause__`` attribute.
    """

    def __init__(self, name: str, original: Exception) -> None:
        self.original: Exception = original
        msg = f"Extension {name!r} raised an error: {original.__class__.__name__}: {original}"
        super().__init__(msg, name=name)


class ExtensionNotFound(ExtensionError):
    """An exception raised when an extension is not found.

    This inherits from :exc:`ExtensionError`

    .. versionchanged:: 1.3
        Made the ``original`` attribute always None.

    Attributes
    ----------
    name: :class:`str`
        The extension that had the error.
    """

    def __init__(self, name: str) -> None:
        msg = f"Extension {name!r} could not be loaded."
        super().__init__(msg, name=name)


class CommandRegistrationError(ClientException):
    """An exception raised when the command can't be added
    because the name is already taken by a different command.

    This inherits from :exc:`disnake.ClientException`

    .. versionadded:: 1.4

    Attributes
    ----------
    name: :class:`str`
        The command name that had the error.
    alias_conflict: :class:`bool`
        Whether the name that conflicts is an alias of the command we try to add.
    """

    def __init__(self, name: str, *, alias_conflict: bool = False) -> None:
        self.name: str = name
        self.alias_conflict: bool = alias_conflict
        type_ = "alias" if alias_conflict else "command"
        super().__init__(f"The {type_} {name} is already an existing command or alias.")


class FlagError(BadArgument):
    """The base exception type for all flag parsing related errors.

    This inherits from :exc:`BadArgument`.

    .. versionadded:: 2.0
    """

    pass


class TooManyFlags(FlagError):
    """An exception raised when a flag has received too many values.

    This inherits from :exc:`FlagError`.

    .. versionadded:: 2.0

    Attributes
    ----------
    flag: :class:`.Flag`
        The flag that received too many values.
    values: List[:class:`str`]
        The values that were passed.
    """

    def __init__(self, flag: Flag, values: List[str]) -> None:
        self.flag: Flag = flag
        self.values: List[str] = values
        super().__init__(
            f"Too many flag values, expected {flag.max_args} but received {len(values)}."
        )


class BadFlagArgument(FlagError):
    """An exception raised when a flag failed to convert a value.

    This inherits from :exc:`FlagError`

    .. versionadded:: 2.0

    Attributes
    ----------
    flag: :class:`.Flag`
        The flag that failed to convert.
    """

    def __init__(self, flag: Flag) -> None:
        self.flag: Flag = flag
        try:
            name = flag.annotation.__name__
        except AttributeError:
            name = flag.annotation.__class__.__name__

        super().__init__(f"Could not convert to {name!r} for flag {flag.name!r}")


class MissingRequiredFlag(FlagError):
    """An exception raised when a required flag was not given.

    This inherits from :exc:`FlagError`

    .. versionadded:: 2.0

    Attributes
    ----------
    flag: :class:`.Flag`
        The required flag that was not found.
    """

    def __init__(self, flag: Flag) -> None:
        self.flag: Flag = flag
        super().__init__(f"Flag {flag.name!r} is required and missing")


class MissingFlagArgument(FlagError):
    """An exception raised when a flag did not get a value.

    This inherits from :exc:`FlagError`

    .. versionadded:: 2.0

    Attributes
    ----------
    flag: :class:`.Flag`
        The flag that did not get a value.
    """

    def __init__(self, flag: Flag) -> None:
        self.flag: Flag = flag
        super().__init__(f"Flag {flag.name!r} does not have an argument")



# SPDX-License-Identifier: MIT

from typing import List, Literal, Optional, TypedDict

from .emoji import Emoji
from .snowflake import Snowflake, SnowflakeList

OnboardingPromptType = Literal[0, 1]


class OnboardingPromptOption(TypedDict):
    id: Snowflake
    title: str
    description: Optional[str]
    emoji: Emoji
    role_ids: SnowflakeList
    channel_ids: SnowflakeList


class OnboardingPrompt(TypedDict):
    id: Snowflake
    title: str
    options: List[OnboardingPromptOption]
    single_select: bool
    required: bool
    in_onboarding: bool
    type: OnboardingPromptType


class Onboarding(TypedDict):
    guild_id: Snowflake
    prompts: List[OnboardingPrompt]
    default_channel_ids: SnowflakeList
    enabled: bool



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Literal, Optional, TypedDict, Union

from typing_extensions import NotRequired

from .appinfo import ApplicationIntegrationType
from .channel import ChannelType
from .components import Component, Modal
from .embed import Embed
from .entitlement import Entitlement
from .i18n import LocalizationDict
from .member import Member, MemberWithUser
from .role import Role
from .snowflake import Snowflake
from .threads import ThreadMetadata
from .user import User

if TYPE_CHECKING:
    from .message import AllowedMentions, Attachment, Message


ApplicationCommandType = Literal[1, 2, 3]

InteractionContextType = Literal[1, 2, 3]  # GUILD, BOT_DM, PRIVATE_CHANNEL


class ApplicationCommand(TypedDict):
    id: Snowflake
    type: NotRequired[ApplicationCommandType]
    application_id: Snowflake
    guild_id: NotRequired[Snowflake]
    name: str
    name_localizations: NotRequired[Optional[LocalizationDict]]
    description: str
    description_localizations: NotRequired[Optional[LocalizationDict]]
    options: NotRequired[List[ApplicationCommandOption]]
    default_member_permissions: NotRequired[Optional[str]]
    dm_permission: NotRequired[Optional[bool]]  # deprecated
    default_permission: NotRequired[bool]  # deprecated
    nsfw: NotRequired[bool]
    integration_types: NotRequired[List[ApplicationIntegrationType]]
    contexts: NotRequired[Optional[List[InteractionContextType]]]
    version: Snowflake


ApplicationCommandOptionType = Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]


class ApplicationCommandOption(TypedDict):
    type: ApplicationCommandOptionType
    name: str
    name_localizations: NotRequired[Optional[LocalizationDict]]
    description: str
    description_localizations: NotRequired[Optional[LocalizationDict]]
    required: NotRequired[bool]
    choices: NotRequired[List[ApplicationCommandOptionChoice]]
    options: NotRequired[List[ApplicationCommandOption]]
    channel_types: NotRequired[List[ChannelType]]
    min_value: NotRequired[float]
    max_value: NotRequired[float]
    min_length: NotRequired[int]
    max_length: NotRequired[int]
    autocomplete: NotRequired[bool]


ApplicationCommandOptionChoiceValue = Union[str, int, float]


class ApplicationCommandOptionChoice(TypedDict):
    name: str
    name_localizations: NotRequired[Optional[LocalizationDict]]
    value: ApplicationCommandOptionChoiceValue


ApplicationCommandPermissionType = Literal[1, 2, 3]


class ApplicationCommandPermissions(TypedDict):
    id: Snowflake
    type: ApplicationCommandPermissionType
    permission: bool


class GuildApplicationCommandPermissions(TypedDict):
    id: Snowflake
    application_id: Snowflake
    guild_id: Snowflake
    permissions: List[ApplicationCommandPermissions]


InteractionType = Literal[1, 2, 3, 4, 5]


class InteractionChannel(TypedDict):
    id: Snowflake
    type: ChannelType
    permissions: str
    name: str

    # only in threads
    thread_metadata: NotRequired[ThreadMetadata]
    parent_id: NotRequired[Snowflake]


class InteractionDataResolved(TypedDict, total=False):
    users: Dict[Snowflake, User]
    members: Dict[Snowflake, Member]
    roles: Dict[Snowflake, Role]
    channels: Dict[Snowflake, InteractionChannel]


class ApplicationCommandInteractionDataResolved(InteractionDataResolved, total=False):
    messages: Dict[Snowflake, Message]
    attachments: Dict[Snowflake, Attachment]


class _ApplicationCommandInteractionDataOption(TypedDict):
    name: str


class _ApplicationCommandInteractionDataOptionSubcommand(_ApplicationCommandInteractionDataOption):
    type: Literal[1, 2]
    options: List[ApplicationCommandInteractionDataOption]


class _ApplicationCommandInteractionDataOptionString(_ApplicationCommandInteractionDataOption):
    type: Literal[3]
    value: str


class _ApplicationCommandInteractionDataOptionInteger(_ApplicationCommandInteractionDataOption):
    type: Literal[4]
    value: int


class _ApplicationCommandInteractionDataOptionBoolean(_ApplicationCommandInteractionDataOption):
    type: Literal[5]
    value: bool


class _ApplicationCommandInteractionDataOptionSnowflake(_ApplicationCommandInteractionDataOption):
    type: Literal[6, 7, 8, 9, 11]
    value: Snowflake


class _ApplicationCommandInteractionDataOptionNumber(_ApplicationCommandInteractionDataOption):
    type: Literal[10]
    value: float


ApplicationCommandInteractionDataOption = Union[
    _ApplicationCommandInteractionDataOptionString,
    _ApplicationCommandInteractionDataOptionInteger,
    _ApplicationCommandInteractionDataOptionSubcommand,
    _ApplicationCommandInteractionDataOptionBoolean,
    _ApplicationCommandInteractionDataOptionSnowflake,
    _ApplicationCommandInteractionDataOptionNumber,
]


class ApplicationCommandInteractionData(TypedDict):
    id: Snowflake
    name: str
    type: ApplicationCommandType
    resolved: NotRequired[ApplicationCommandInteractionDataResolved]
    options: NotRequired[List[ApplicationCommandInteractionDataOption]]
    # this is the guild the command is registered to, not the guild the command was invoked in (see interaction.guild_id)
    guild_id: NotRequired[Snowflake]
    target_id: NotRequired[Snowflake]


## Interaction components


class _BaseComponentInteractionData(TypedDict):
    custom_id: str


class _BaseSnowflakeComponentInteractionData(_BaseComponentInteractionData):
    values: List[Snowflake]
    resolved: NotRequired[InteractionDataResolved]


### Message interaction components


class MessageComponentInteractionButtonData(_BaseComponentInteractionData):
    component_type: Literal[2]


class MessageComponentInteractionStringSelectData(_BaseComponentInteractionData):
    component_type: Literal[3]
    values: List[str]


class MessageComponentInteractionUserSelectData(_BaseSnowflakeComponentInteractionData):
    component_type: Literal[5]


class MessageComponentInteractionRoleSelectData(_BaseSnowflakeComponentInteractionData):
    component_type: Literal[6]


class MessageComponentInteractionMentionableSelectData(_BaseSnowflakeComponentInteractionData):
    component_type: Literal[7]


class MessageComponentInteractionChannelSelectData(_BaseSnowflakeComponentInteractionData):
    component_type: Literal[8]


MessageComponentInteractionData = Union[
    MessageComponentInteractionButtonData,
    MessageComponentInteractionStringSelectData,
    MessageComponentInteractionUserSelectData,
    MessageComponentInteractionRoleSelectData,
    MessageComponentInteractionMentionableSelectData,
    MessageComponentInteractionChannelSelectData,
]


### Modal interaction components


# TODO: add other select types
class ModalInteractionStringSelectData(_BaseComponentInteractionData):
    type: Literal[3]
    values: List[str]


class ModalInteractionTextInputData(_BaseComponentInteractionData):
    type: Literal[4]
    value: str


ModalInteractionComponentData = Union[
    ModalInteractionStringSelectData,
    ModalInteractionTextInputData,
]


class ModalInteractionActionRow(TypedDict):
    type: Literal[1]
    components: List[ModalInteractionComponentData]


class ModalInteractionData(TypedDict):
    custom_id: str
    components: List[ModalInteractionActionRow]


## Interactions


# keys are stringified ApplicationInstallType's
AuthorizingIntegrationOwners = Dict[str, Snowflake]


# base type for *all* interactions
class _BaseInteraction(TypedDict):
    id: Snowflake
    application_id: Snowflake
    token: str
    version: Literal[1]
    app_permissions: str


# common properties in non-ping interactions
class _BaseUserInteraction(_BaseInteraction):
    # the docs specify `channel_id` and 'channel` as optional,
    # but they're assumed to always exist on non-ping interactions
    channel_id: Snowflake
    channel: InteractionChannel
    locale: str
    guild_id: NotRequired[Snowflake]
    guild_locale: NotRequired[str]
    entitlements: NotRequired[List[Entitlement]]
    authorizing_integration_owners: NotRequired[AuthorizingIntegrationOwners]
    context: NotRequired[InteractionContextType]
    # one of these two will always exist, according to docs
    member: NotRequired[MemberWithUser]
    user: NotRequired[User]


class PingInteraction(_BaseInteraction):
    type: Literal[1]


class ApplicationCommandInteraction(_BaseUserInteraction):
    type: Literal[2, 4]
    data: ApplicationCommandInteractionData


class MessageInteraction(_BaseUserInteraction):
    type: Literal[3]
    data: MessageComponentInteractionData
    message: Message


class ModalInteraction(_BaseUserInteraction):
    type: Literal[5]
    data: ModalInteractionData
    message: NotRequired[Message]


Interaction = Union[
    ApplicationCommandInteraction,
    MessageInteraction,
    ModalInteraction,
]

BaseInteraction = Union[Interaction, PingInteraction]


class InteractionApplicationCommandCallbackData(TypedDict, total=False):
    tts: bool
    content: str
    embeds: List[Embed]
    allowed_mentions: AllowedMentions
    flags: int
    components: List[Component]
    attachments: List[Attachment]


class InteractionAutocompleteCallbackData(TypedDict):
    choices: List[ApplicationCommandOptionChoice]


InteractionResponseType = Literal[1, 4, 5, 6, 7, 10]

InteractionCallbackData = Union[
    InteractionApplicationCommandCallbackData,
    InteractionAutocompleteCallbackData,
    Modal,
]


class InteractionResponse(TypedDict):
    type: InteractionResponseType
    data: NotRequired[InteractionCallbackData]


class InteractionMessageReference(TypedDict):
    id: Snowflake
    type: InteractionType
    name: str
    user: User
    member: NotRequired[Member]


class _BaseInteractionMetadata(TypedDict):
    id: Snowflake
    type: InteractionType
    user: User
    authorizing_integration_owners: AuthorizingIntegrationOwners
    original_response_message_id: NotRequired[Snowflake]  # only on followups


class ApplicationCommandInteractionMetadata(_BaseInteractionMetadata):
    target_user: NotRequired[User]  # only on user command interactions
    target_message_id: NotRequired[Snowflake]  # only on message command interactions


class MessageComponentInteractionMetadata(_BaseInteractionMetadata):
    interacted_message_id: Snowflake


class ModalInteractionMetadata(_BaseInteractionMetadata):
    triggering_interaction_metadata: Union[
        ApplicationCommandInteractionMetadata,
        MessageComponentInteractionMetadata,
    ]


InteractionMetadata = Union[
    ApplicationCommandInteractionMetadata,
    MessageComponentInteractionMetadata,
    ModalInteractionMetadata,
]


class EditApplicationCommand(TypedDict):
    name: str
    name_localizations: NotRequired[Optional[LocalizationDict]]
    description: NotRequired[str]
    description_localizations: NotRequired[Optional[LocalizationDict]]
    options: NotRequired[Optional[List[ApplicationCommandOption]]]
    default_member_permissions: NotRequired[Optional[str]]
    dm_permission: NotRequired[bool]  # deprecated
    default_permission: NotRequired[bool]  # deprecated
    nsfw: NotRequired[bool]
    integration_types: NotRequired[Optional[List[ApplicationIntegrationType]]]
    contexts: NotRequired[Optional[List[InteractionContextType]]]
    # n.b. this cannot be changed
    type: NotRequired[ApplicationCommandType]



# SPDX-License-Identifier: MIT

"""disnake.ui
~~~~~~~~~~~

Bot UI Kit helper for the Discord API

:copyright: (c) 2015-2021 Rapptz, 2021-present Disnake Development
:license: MIT, see LICENSE for more details.

"""
from .action_row import *
from .button import *
from .item import *
from .modal import *
from .select import *
from .text_input import *
from .view import *



# SPDX-License-Identifier: MIT

from typing import Literal, TypedDict

from typing_extensions import NotRequired

from .snowflake import Snowflake

EntitlementType = Literal[1, 2, 3, 4, 5, 6, 7, 8]


class Entitlement(TypedDict):
    id: Snowflake
    sku_id: Snowflake
    user_id: NotRequired[Snowflake]
    guild_id: NotRequired[Snowflake]
    application_id: Snowflake
    type: EntitlementType
    deleted: bool
    starts_at: NotRequired[str]
    ends_at: NotRequired[str]
    consumed: NotRequired[bool]



# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import collections
import collections.abc
import inspect
import logging
import sys
import traceback
import warnings
from typing import TYPE_CHECKING, Any, Callable, Iterable, List, Optional, Type, TypeVar, Union

import disnake

from . import errors
from .common_bot_base import CommonBotBase
from .context import Context
from .core import GroupMixin
from .custom_warnings import MessageContentPrefixWarning
from .help import DefaultHelpCommand, HelpCommand
from .view import StringView

if TYPE_CHECKING:
    from typing_extensions import Self

    from disnake.message import Message

    from ._types import Check, CoroFunc, MaybeCoro

__all__ = (
    "when_mentioned",
    "when_mentioned_or",
    "BotBase",
)

MISSING: Any = disnake.utils.MISSING

T = TypeVar("T")
CFT = TypeVar("CFT", bound="CoroFunc")
CXT = TypeVar("CXT", bound="Context")

PrefixType = Union[str, Iterable[str]]

_log = logging.getLogger(__name__)


def when_mentioned(bot: BotBase, msg: Message) -> List[str]:
    """A callable that implements a command prefix equivalent to being mentioned.

    These are meant to be passed into the :attr:`.Bot.command_prefix` attribute.
    """
    # bot.user will never be None when this is called
    return [f"<@{bot.user.id}> ", f"<@!{bot.user.id}> "]  # type: ignore


def when_mentioned_or(*prefixes: str) -> Callable[[BotBase, Message], List[str]]:
    """A callable that implements when mentioned or other prefixes provided.

    These are meant to be passed into the :attr:`.Bot.command_prefix` attribute.

    Example
    -------
    .. code-block:: python3

        bot = commands.Bot(command_prefix=commands.when_mentioned_or('!'))


    .. note::

        This callable returns another callable, so if this is done inside a custom
        callable, you must call the returned callable, for example:

        .. code-block:: python3

            async def get_prefix(bot, message):
                extras = await prefixes_for(message.guild) # returns a list
                return commands.when_mentioned_or(*extras)(bot, message)


    See Also
    --------
    :func:`.when_mentioned`
    """

    def inner(bot, msg):
        r = list(prefixes)
        r = when_mentioned(bot, msg) + r
        return r

    return inner


def _is_submodule(parent: str, child: str) -> bool:
    return parent == child or child.startswith(parent + ".")


class _DefaultRepr:
    def __repr__(self) -> str:
        return "<default-help-command>"


_default: Any = _DefaultRepr()


class BotBase(CommonBotBase, GroupMixin):
    def __init__(
        self,
        command_prefix: Optional[
            Union[PrefixType, Callable[[Self, Message], MaybeCoro[PrefixType]]]
        ] = None,
        help_command: Optional[HelpCommand] = _default,
        description: Optional[str] = None,
        *,
        strip_after_prefix: bool = False,
        **options: Any,
    ) -> None:
        super().__init__(**options)

        if not isinstance(self, disnake.Client):
            raise RuntimeError("BotBase mixin must be used with disnake.Client")  # noqa: TRY004

        alternative = (
            "AutoShardedInteractionBot"
            if isinstance(self, disnake.AutoShardedClient)
            else "InteractionBot"
        )
        if command_prefix is None:
            disnake.utils.warn_deprecated(
                "Using `command_prefix=None` is deprecated and will result in "
                "an error in future versions. "
                f"If you don't need any prefix functionality, consider using {alternative}.",
                stacklevel=2,
            )
        elif (
            # note: no need to check for empty iterables,
            # as they won't be allowed by `get_prefix`
            command_prefix is not when_mentioned
            and not self.intents.message_content
        ):
            warnings.warn(
                "Message Content intent is not enabled and a prefix is configured. "
                "This may cause limited functionality for prefix commands. "
                "If you want prefix commands, pass an intents object with message_content set to True. "
                f"If you don't need any prefix functionality, consider using {alternative}. "
                "Alternatively, set prefix to disnake.ext.commands.when_mentioned to silence this warning.",
                MessageContentPrefixWarning,
                stacklevel=2,
            )

        self.command_prefix = command_prefix

        self._checks: List[Check] = []
        self._check_once: List[Check] = []

        self._before_invoke: Optional[CoroFunc] = None
        self._after_invoke: Optional[CoroFunc] = None

        self._help_command: Optional[HelpCommand] = None
        self.description: str = inspect.cleandoc(description) if description else ""
        self.strip_after_prefix: bool = strip_after_prefix

        if help_command is _default:
            self.help_command = DefaultHelpCommand()
        else:
            self.help_command = help_command

    # internal helpers

    async def on_command_error(self, context: Context, exception: errors.CommandError) -> None:
        """|coro|

        The default command error handler provided by the bot.

        This is for text commands only, and doesn't apply to application commands.

        By default this prints to :data:`sys.stderr` however it could be
        overridden to have a different implementation.

        This only fires if you do not specify any listeners for command error.
        """
        if self.extra_events.get("on_command_error", None):
            return

        command = context.command
        if command and command.has_error_handler():
            return

        cog = context.cog
        if cog and cog.has_error_handler():
            return

        print(f"Ignoring exception in command {context.command}:", file=sys.stderr)
        traceback.print_exception(
            type(exception), exception, exception.__traceback__, file=sys.stderr
        )

    # global check registration

    def add_check(
        self,
        func: Check,
        *,
        call_once: bool = False,
    ) -> None:
        """Adds a global check to the bot.

        This is for text commands only, and doesn't apply to application commands.

        This is the non-decorator interface to :meth:`.check` and :meth:`.check_once`.

        Parameters
        ----------
        func
            The function that was used as a global check.
        call_once: :class:`bool`
            If the function should only be called once per
            :meth:`.invoke` call.
        """
        if call_once:
            self._check_once.append(func)
        else:
            self._checks.append(func)

    def remove_check(
        self,
        func: Check,
        *,
        call_once: bool = False,
    ) -> None:
        """Removes a global check from the bot.

        This is for text commands only, and doesn't apply to application commands.

        This function is idempotent and will not raise an exception
        if the function is not in the global checks.

        Parameters
        ----------
        func
            The function to remove from the global checks.
        call_once: :class:`bool`
            If the function was added with ``call_once=True`` in
            the :meth:`.Bot.add_check` call or using :meth:`.check_once`.
        """
        check_list = self._check_once if call_once else self._checks
        try:
            check_list.remove(func)
        except ValueError:
            pass

    def check(self, func: T) -> T:
        """A decorator that adds a global check to the bot.

        This is for text commands only, and doesn't apply to application commands.

        A global check is similar to a :func:`.check` that is applied
        on a per command basis except it is run before any command checks
        have been verified and applies to every command the bot has.

        .. note::

            This function can either be a regular function or a coroutine.

        Similar to a command :func:`.check`\\, this takes a single parameter
        of type :class:`.Context` and can only raise exceptions inherited from
        :exc:`.CommandError`.

        Example
        -------
        .. code-block:: python3

            @bot.check
            def check_commands(ctx):
                return ctx.command.qualified_name in allowed_commands

        """
        # T was used instead of Check to ensure the type matches on return
        self.add_check(func)  # type: ignore
        return func

    def check_once(self, func: CFT) -> CFT:
        """A decorator that adds a "call once" global check to the bot.

        This is for text commands only, and doesn't apply to application commands.

        Unlike regular global checks, this one is called only once
        per :meth:`.invoke` call.

        Regular global checks are called whenever a command is called
        or :meth:`.Command.can_run` is called. This type of check
        bypasses that and ensures that it's called only once, even inside
        the default help command.

        .. note::

            When using this function the :class:`.Context` sent to a group subcommand
            may only parse the parent command and not the subcommands due to it
            being invoked once per :meth:`.Bot.invoke` call.

        .. note::

            This function can either be a regular function or a coroutine.

        Similar to a command :func:`.check`\\, this takes a single parameter
        of type :class:`.Context` and can only raise exceptions inherited from
        :exc:`.CommandError`.

        Example
        -------
        .. code-block:: python3

            @bot.check_once
            def whitelist(ctx):
                return ctx.message.author.id in my_whitelist

        """
        self.add_check(func, call_once=True)
        return func

    async def can_run(self, ctx: Context, *, call_once: bool = False) -> bool:
        data = self._check_once if call_once else self._checks

        if len(data) == 0:
            return True

        # type-checker doesn't distinguish between functions and methods
        return await disnake.utils.async_all(f(ctx) for f in data)  # type: ignore

    def before_invoke(self, coro: CFT) -> CFT:
        """A decorator that registers a coroutine as a pre-invoke hook.

        This is for text commands only, and doesn't apply to application commands.

        A pre-invoke hook is called directly before the command is
        called. This makes it a useful function to set up database
        connections or any type of set up required.

        This pre-invoke hook takes a sole parameter, a :class:`.Context`.

        .. note::

            The :meth:`~.Bot.before_invoke` and :meth:`~.Bot.after_invoke` hooks are
            only called if all checks and argument parsing procedures pass
            without error. If any check or argument parsing procedures fail
            then the hooks are not called.

        Parameters
        ----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the pre-invoke hook.

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("The pre-invoke hook must be a coroutine.")

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro: CFT) -> CFT:
        """A decorator that registers a coroutine as a post-invoke hook.

        This is for text commands only, and doesn't apply to application commands.

        A post-invoke hook is called directly after the command is
        called. This makes it a useful function to clean-up database
        connections or any type of clean up required.

        This post-invoke hook takes a sole parameter, a :class:`.Context`.

        .. note::

            Similar to :meth:`~.Bot.before_invoke`\\, this is not called unless
            checks and argument parsing procedures succeed. This hook is,
            however, **always** called regardless of the internal command
            callback raising an error (i.e. :exc:`.CommandInvokeError`\\).
            This makes it ideal for clean-up scenarios.

        Parameters
        ----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the post-invoke hook.

        Raises
        ------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError("The post-invoke hook must be a coroutine.")

        self._after_invoke = coro
        return coro

    # extensions

    def _remove_module_references(self, name: str) -> None:
        super()._remove_module_references(name)
        # remove all the commands from the module
        for cmd in self.all_commands.copy().values():
            if cmd.module and _is_submodule(name, cmd.module):
                if isinstance(cmd, GroupMixin):
                    cmd.recursively_remove_all_commands()
                self.remove_command(cmd.name)

    # help command stuff

    @property
    def help_command(self) -> Optional[HelpCommand]:
        return self._help_command

    @help_command.setter
    def help_command(self, value: Optional[HelpCommand]) -> None:
        if value is not None and not isinstance(value, HelpCommand):
            raise TypeError("help_command must be a subclass of HelpCommand or None")

        if self._help_command is not None:
            self._help_command._remove_from_bot(self)

        self._help_command = value

        if value is not None:
            value._add_to_bot(self)

    # command processing

    async def get_prefix(self, message: Message) -> Optional[Union[List[str], str]]:
        """|coro|

        Retrieves the prefix the bot is listening to
        with the message as a context.

        Parameters
        ----------
        message: :class:`disnake.Message`
            The message context to get the prefix of.

        Returns
        -------
        Optional[Union[List[:class:`str`], :class:`str`]]
            A list of prefixes or a single prefix that the bot is
            listening for. None if the bot isn't listening for prefixes.
        """
        ret = self.command_prefix
        if callable(ret):
            ret = await disnake.utils.maybe_coroutine(ret, self, message)

        if ret is None:
            return None

        if not isinstance(ret, str):
            try:
                ret = list(ret)
            except TypeError:
                # It's possible that a generator raised this exception.  Don't
                # replace it with our own error if that's the case.
                if isinstance(ret, collections.abc.Iterable):
                    raise

                raise TypeError(
                    "command_prefix must be plain string, iterable of strings, or callable "
                    f"returning either of these, not {ret.__class__.__name__}"
                ) from None

            if not ret:
                raise ValueError("Iterable command_prefix must contain at least one prefix")

        return ret

    async def get_context(self, message: Message, *, cls: Type[CXT] = Context) -> CXT:
        """|coro|

        Returns the invocation context from the message.

        This is a more low-level counter-part for :meth:`.process_commands`
        to allow users more fine grained control over the processing.

        The returned context is not guaranteed to be a valid invocation
        context, :attr:`.Context.valid` must be checked to make sure it is.
        If the context is not valid then it is not a valid candidate to be
        invoked under :meth:`~.Bot.invoke`.

        Parameters
        ----------
        message: :class:`disnake.Message`
            The message to get the invocation context from.
        cls
            The factory class that will be used to create the context.
            By default, this is :class:`.Context`. Should a custom
            class be provided, it must be similar enough to :class:`.Context`\'s
            interface.

        Returns
        -------
        :class:`.Context`
            The invocation context. The type of this can change via the
            ``cls`` parameter.
        """
        view = StringView(message.content)
        ctx = cls(prefix=None, view=view, bot=self, message=message)

        if message.author.id == self.user.id:  # type: ignore
            return ctx

        prefix = await self.get_prefix(message)
        invoked_prefix = prefix

        if prefix is None:
            return ctx
        elif isinstance(prefix, str):
            if not view.skip_string(prefix):
                return ctx
        else:
            try:
                # if the context class' __init__ consumes something from the view this
                # will be wrong.  That seems unreasonable though.
                if message.content.startswith(tuple(prefix)):
                    invoked_prefix = disnake.utils.find(view.skip_string, prefix)
                else:
                    return ctx

            except TypeError:
                if not isinstance(prefix, list):
                    raise TypeError(
                        "get_prefix must return either a string or a list of string, "
                        f"not {prefix.__class__.__name__}"
                    ) from None

                # It's possible a bad command_prefix got us here.
                for value in prefix:
                    if not isinstance(value, str):
                        raise TypeError(
                            "Iterable command_prefix or list returned from get_prefix must "
                            f"contain only strings, not {value.__class__.__name__}"
                        ) from None

                # Getting here shouldn't happen
                raise

        if self.strip_after_prefix:
            view.skip_ws()

        invoker = view.get_word()
        ctx.invoked_with = invoker
        # type-checker fails to narrow invoked_prefix type.
        ctx.prefix = invoked_prefix  # type: ignore
        ctx.command = self.all_commands.get(invoker)
        return ctx

    async def invoke(self, ctx: Context) -> None:
        """|coro|

        Invokes the command given under the invocation context and
        handles all the internal event dispatch mechanisms.

        Parameters
        ----------
        ctx: :class:`.Context`
            The invocation context to invoke.
        """
        if ctx.command is not None:
            self.dispatch("command", ctx)
            try:
                if await self.can_run(ctx, call_once=True):
                    await ctx.command.invoke(ctx)
                else:
                    raise errors.CheckFailure("The global check once functions failed.")
            except errors.CommandError as exc:
                await ctx.command.dispatch_error(ctx, exc)
            else:
                self.dispatch("command_completion", ctx)
        elif ctx.invoked_with:
            exc = errors.CommandNotFound(f'Command "{ctx.invoked_with}" is not found')
            self.dispatch("command_error", ctx, exc)

    async def process_commands(self, message: Message) -> None:
        """|coro|

        This function processes the commands that have been registered
        to the bot and other groups. Without this coroutine, none of the
        commands will be triggered.

        By default, this coroutine is called inside the :func:`.on_message`
        event. If you choose to override the :func:`.on_message` event, then
        you should invoke this coroutine as well.

        This is built using other low level tools, and is equivalent to a
        call to :meth:`~.Bot.get_context` followed by a call to :meth:`~.Bot.invoke`.

        This also checks if the message's author is a bot and doesn't
        call :meth:`~.Bot.get_context` or :meth:`~.Bot.invoke` if so.

        Parameters
        ----------
        message: :class:`disnake.Message`
            The message to process commands for.
        """
        if message.author.bot:
            return

        ctx = await self.get_context(message)
        await self.invoke(ctx)

    async def on_message(self, message) -> None:
        await self.process_commands(message)



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, cast

from . import utils
from .asset import Asset
from .flags import ApplicationFlags
from .permissions import Permissions

if TYPE_CHECKING:
    from .guild import Guild
    from .state import ConnectionState
    from .types.appinfo import (
        AppInfo as AppInfoPayload,
        ApplicationIntegrationType as ApplicationIntegrationTypeLiteral,
        ApplicationIntegrationTypeConfiguration as ApplicationIntegrationTypeConfigurationPayload,
        InstallParams as InstallParamsPayload,
        PartialAppInfo as PartialAppInfoPayload,
        Team as TeamPayload,
    )
    from .user import User

__all__ = (
    "AppInfo",
    "PartialAppInfo",
    "InstallParams",
    "InstallTypeConfiguration",
)


class InstallParams:
    """Represents the installation parameters for the application, provided by Discord.

    .. versionadded:: 2.5

    Attributes
    ----------
    scopes: List[:class:`str`]
        The scopes requested by the application.
    permissions: :class:`Permissions`
        The permissions requested for the bot role.
    """

    __slots__ = (
        "_app_id",
        "_install_type",
        "scopes",
        "permissions",
    )

    def __init__(
        self,
        data: InstallParamsPayload,
        parent: AppInfo,
        *,
        install_type: Optional[ApplicationIntegrationTypeLiteral] = None,
    ) -> None:
        self._app_id = parent.id
        self._install_type: Optional[ApplicationIntegrationTypeLiteral] = install_type
        self.scopes = data["scopes"]
        self.permissions = Permissions(int(data["permissions"]))

    def __repr__(self) -> str:
        return f"<InstallParams scopes={self.scopes!r} permissions={self.permissions!r}>"

    def to_url(self) -> str:
        """Returns a string that can be used to install this application.

        Returns
        -------
        :class:`str`
            The invite url.
        """
        return utils.oauth_url(
            self._app_id,
            scopes=self.scopes,
            permissions=self.permissions,
            integration_type=(
                self._install_type if self._install_type is not None else utils.MISSING
            ),
        )


class InstallTypeConfiguration:
    """Represents the configuration for a particular application installation type.

    .. versionadded:: 2.10

    Attributes
    ----------
    install_params: Optional[:class:`InstallParams`]
        The parameters for this installation type.
    """

    __slots__ = ("install_params",)

    def __init__(
        self,
        data: ApplicationIntegrationTypeConfigurationPayload,
        *,
        parent: AppInfo,
        install_type: ApplicationIntegrationTypeLiteral,
    ) -> None:
        self.install_params: Optional[InstallParams] = (
            InstallParams(install_params, parent=parent, install_type=install_type)
            if (install_params := data.get("oauth2_install_params"))
            else None
        )


class AppInfo:
    """Represents the application info for the bot provided by Discord.

    Attributes
    ----------
    id: :class:`int`
        The application's ID.
    name: :class:`str`
        The application's name.
    owner: :class:`User`
        The application's owner.
    team: Optional[:class:`Team`]
        The application's team.

        .. versionadded:: 1.3

    description: :class:`str`
        The application's description.
    bot_public: :class:`bool`
        Whether the bot can be invited by anyone or if it is locked
        to the application owner.
    bot_require_code_grant: :class:`bool`
        Whether the bot requires the completion of the full oauth2 code
        grant flow to join.
    rpc_origins: Optional[List[:class:`str`]]
        A list of RPC origin URLs, if RPC is enabled.
    verify_key: :class:`str`
        The hex encoded key for verification in interactions and the
        GameSDK's :ddocs:`GetTicket <game-sdk/applications#getticket>`.

        .. versionadded:: 1.3

    guild_id: Optional[:class:`int`]
        The ID of the guild associated with the application, if any.

        .. versionadded:: 1.3

    primary_sku_id: Optional[:class:`int`]
        If this application is a game sold on Discord,
        this field will be the ID of the "Game SKU" that is created,
        if it exists.

        .. versionadded:: 1.3

    slug: Optional[:class:`str`]
        If this application is a game sold on Discord,
        this field will be the URL slug that links to the store page.

        .. versionadded:: 1.3

    terms_of_service_url: Optional[:class:`str`]
        The application's terms of service URL, if set.

        .. versionadded:: 2.0

    privacy_policy_url: Optional[:class:`str`]
        The application's privacy policy URL, if set.

        .. versionadded:: 2.0

    flags: Optional[:class:`ApplicationFlags`]
        The application's public flags.

        .. versionadded:: 2.3

    tags: Optional[List[:class:`str`]]
        The application's tags.

        .. versionadded:: 2.5

    install_params: Optional[:class:`InstallParams`]
        The installation parameters for this application.

        See also :attr:`guild_install_type_config`/:attr:`user_install_type_config`
        for installation type-specific configuration.

        .. versionadded:: 2.5

    custom_install_url: Optional[:class:`str`]
        The custom installation url for this application.

        .. versionadded:: 2.5
    role_connections_verification_url: Optional[:class:`str`]
        The application's role connection verification entry point,
        which when configured will render the app as a verification method
        in the guild role verification configuration.

        .. versionadded:: 2.8
    approximate_guild_count: :class:`int`
        The approximate number of guilds the application is installed to.

        .. versionadded:: 2.10
    approximate_user_install_count: :class:`int`
        The approximate number of users that have installed the application
        (for user-installable apps).

        .. versionadded:: 2.10
    """

    __slots__ = (
        "_state",
        "description",
        "id",
        "name",
        "rpc_origins",
        "bot_public",
        "bot_require_code_grant",
        "owner",
        "_icon",
        "_summary",
        "verify_key",
        "team",
        "guild_id",
        "primary_sku_id",
        "slug",
        "_cover_image",
        "terms_of_service_url",
        "privacy_policy_url",
        "flags",
        "tags",
        "install_params",
        "custom_install_url",
        "role_connections_verification_url",
        "approximate_guild_count",
        "approximate_user_install_count",
        "_install_types_config",
    )

    def __init__(self, state: ConnectionState, data: AppInfoPayload) -> None:
        from .team import Team

        self._state: ConnectionState = state
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self.description: str = data["description"]
        self._icon: Optional[str] = data["icon"]
        self.rpc_origins: List[str] = data["rpc_origins"]
        self.bot_public: bool = data["bot_public"]
        self.bot_require_code_grant: bool = data["bot_require_code_grant"]
        self.owner: User = state.create_user(data["owner"])

        team: Optional[TeamPayload] = data.get("team")
        self.team: Optional[Team] = Team(state, team) if team else None

        self._summary: str = data.get("summary", "")
        self.verify_key: str = data["verify_key"]

        self.guild_id: Optional[int] = utils._get_as_snowflake(data, "guild_id")

        self.primary_sku_id: Optional[int] = utils._get_as_snowflake(data, "primary_sku_id")
        self.slug: Optional[str] = data.get("slug")
        self._cover_image: Optional[str] = data.get("cover_image")
        self.terms_of_service_url: Optional[str] = data.get("terms_of_service_url")
        self.privacy_policy_url: Optional[str] = data.get("privacy_policy_url")

        flags: Optional[int] = data.get("flags")
        self.flags: Optional[ApplicationFlags] = (
            ApplicationFlags._from_value(flags) if flags is not None else None
        )
        self.tags: Optional[List[str]] = data.get("tags")
        self.install_params: Optional[InstallParams] = (
            InstallParams(data["install_params"], parent=self) if "install_params" in data else None
        )
        self.custom_install_url: Optional[str] = data.get("custom_install_url")
        self.role_connections_verification_url: Optional[str] = data.get(
            "role_connections_verification_url"
        )
        self.approximate_guild_count: int = data.get("approximate_guild_count", 0)
        self.approximate_user_install_count: int = data.get("approximate_user_install_count", 0)

        # this is a bit of a mess, but there's no better way to expose this data for now
        self._install_types_config: Dict[
            ApplicationIntegrationTypeLiteral, InstallTypeConfiguration
        ] = {}
        for type_str, config in (data.get("integration_types_config") or {}).items():
            install_type = cast("ApplicationIntegrationTypeLiteral", int(type_str))
            self._install_types_config[install_type] = InstallTypeConfiguration(
                config or {},
                parent=self,
                install_type=install_type,
            )

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id} name={self.name!r} "
            f"description={self.description!r} public={self.bot_public} "
            f"owner={self.owner!r}>"
        )

    @property
    def icon(self) -> Optional[Asset]:
        """Optional[:class:`.Asset`]: Retrieves the application's icon asset, if any."""
        if self._icon is None:
            return None
        return Asset._from_icon(self._state, self.id, self._icon, path="app")

    @property
    def cover_image(self) -> Optional[Asset]:
        """Optional[:class:`.Asset`]: Retrieves the cover image on a store embed, if any.

        This is only available if the application is a game sold on Discord.
        """
        if self._cover_image is None:
            return None
        return Asset._from_cover_image(self._state, self.id, self._cover_image)

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild associated with the application, if any.

        .. versionadded:: 1.3
        """
        return self._state._get_guild(self.guild_id)

    @property
    def summary(self) -> str:
        """:class:`str`: If this application is a game sold on Discord,
        this field will be the summary field for the store page of its primary SKU.

        .. versionadded:: 1.3

        .. deprecated:: 2.5

            This field is deprecated by discord and is now always blank. Consider using :attr:`.description` instead.
        """
        utils.warn_deprecated(
            "summary is deprecated and will be removed in a future version. Consider using description instead.",
            stacklevel=2,
        )
        return self._summary

    @property
    def guild_install_type_config(self) -> Optional[InstallTypeConfiguration]:
        """Optional[:class:`InstallTypeConfiguration`]: The guild installation parameters for
        this application. If this application cannot be installed to guilds, returns ``None``.

        .. versionadded:: 2.10
        """
        return self._install_types_config.get(0)

    @property
    def user_install_type_config(self) -> Optional[InstallTypeConfiguration]:
        """Optional[:class:`InstallTypeConfiguration`]: The user installation parameters for
        this application. If this application cannot be installed to users, returns ``None``.

        .. versionadded:: 2.10
        """
        return self._install_types_config.get(1)


class PartialAppInfo:
    """Represents a partial AppInfo given by :func:`~disnake.abc.GuildChannel.create_invite`.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The application's ID.
    name: :class:`str`
        The application's name.
    description: :class:`str`
        The application's description.
    rpc_origins: Optional[List[:class:`str`]]
        A list of RPC origin URLs, if RPC is enabled.
    verify_key: :class:`str`
        The hex encoded key for verification in interactions and the
        GameSDK's :ddocs:`GetTicket <game-sdk/applications#getticket>`.
    terms_of_service_url: Optional[:class:`str`]
        The application's terms of service URL, if set.
    privacy_policy_url: Optional[:class:`str`]
        The application's privacy policy URL, if set.
    """

    __slots__ = (
        "_state",
        "id",
        "name",
        "description",
        "rpc_origins",
        "_summary",
        "verify_key",
        "terms_of_service_url",
        "privacy_policy_url",
        "_icon",
    )

    def __init__(self, *, state: ConnectionState, data: PartialAppInfoPayload) -> None:
        self._state: ConnectionState = state
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self._icon: Optional[str] = data.get("icon")
        self.description: str = data["description"]
        self.rpc_origins: Optional[List[str]] = data.get("rpc_origins")
        self._summary: str = data.get("summary", "")
        self.verify_key: str = data["verify_key"]
        self.terms_of_service_url: Optional[str] = data.get("terms_of_service_url")
        self.privacy_policy_url: Optional[str] = data.get("privacy_policy_url")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name!r} description={self.description!r}>"

    @property
    def icon(self) -> Optional[Asset]:
        """Optional[:class:`.Asset`]: Retrieves the application's icon asset, if any."""
        if self._icon is None:
            return None
        return Asset._from_icon(self._state, self.id, self._icon, path="app")

    @property
    def summary(self) -> str:
        """:class:`str`: If this application is a game sold on Discord,
        this field will be the summary field for the store page of its primary SKU.

        .. deprecated:: 2.5

            This field is deprecated by discord and is now always blank. Consider using :attr:`.description` instead.
        """
        utils.warn_deprecated(
            "summary is deprecated and will be removed in a future version. Consider using description instead.",
            stacklevel=2,
        )
        return self._summary



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Optional, Tuple

from ..components import TextInput as TextInputComponent
from ..enums import ComponentType, TextInputStyle
from ..utils import MISSING
from .item import WrappedComponent

__all__ = ("TextInput",)


class TextInput(WrappedComponent):
    """Represents a UI text input.

    This can only be used in a :class:`~.ui.Modal`.

    .. versionadded:: 2.4

    Parameters
    ----------
    label: :class:`str`
        The label of the text input.
    custom_id: :class:`str`
        The ID of the text input that gets received during an interaction.
    style: :class:`.TextInputStyle`
        The style of the text input.
    placeholder: Optional[:class:`str`]
        The placeholder text that is shown if nothing is entered.
    value: Optional[:class:`str`]
        The pre-filled value of the text input.
    required: :class:`bool`
        Whether the text input is required. Defaults to ``True``.
    min_length: Optional[:class:`int`]
        The minimum length of the text input.
    max_length: Optional[:class:`int`]
        The maximum length of the text input.
    """

    __repr_attributes__: Tuple[str, ...] = (
        "style",
        "label",
        "custom_id",
        "placeholder",
        "value",
        "required",
        "min_length",
        "max_length",
    )
    # We have to set this to MISSING in order to overwrite the abstract property from WrappedComponent
    _underlying: TextInputComponent = MISSING

    def __init__(
        self,
        *,
        label: str,
        custom_id: str,
        style: TextInputStyle = TextInputStyle.short,
        placeholder: Optional[str] = None,
        value: Optional[str] = None,
        required: bool = True,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
    ) -> None:
        self._underlying = TextInputComponent._raw_construct(
            type=ComponentType.text_input,
            style=style,
            label=label,
            custom_id=custom_id,
            placeholder=placeholder,
            value=value,
            required=required,
            min_length=min_length,
            max_length=max_length,
        )

    @property
    def width(self) -> int:
        return 5

    @property
    def style(self) -> TextInputStyle:
        """:class:`.TextInputStyle`: The style of the text input."""
        return self._underlying.style

    @style.setter
    def style(self, value: TextInputStyle) -> None:
        self._underlying.style = value

    @property
    def label(self) -> str:
        """:class:`str`: The label of the text input."""
        return self._underlying.label  # type: ignore

    @label.setter
    def label(self, value: str) -> None:
        self._underlying.label = value

    @property
    def custom_id(self) -> str:
        """:class:`str`: The ID of the text input that gets received during an interaction."""
        return self._underlying.custom_id

    @custom_id.setter
    def custom_id(self, value: str) -> None:
        self._underlying.custom_id = value

    @property
    def placeholder(self) -> Optional[str]:
        """Optional[:class:`str`]: The placeholder text that is shown if nothing is entered."""
        return self._underlying.placeholder

    @placeholder.setter
    def placeholder(self, value: Optional[str]) -> None:
        self._underlying.placeholder = value

    @property
    def value(self) -> Optional[str]:
        """Optional[:class:`str`]: The pre-filled text of the text input."""
        return self._underlying.value

    @value.setter
    def value(self, value: Optional[str]) -> None:
        self._underlying.value = value

    @property
    def required(self) -> bool:
        """:class:`bool`: Whether the text input is required."""
        return self._underlying.required

    @required.setter
    def required(self, value: bool) -> None:
        self._underlying.required = value

    @property
    def min_length(self) -> Optional[int]:
        """Optional[:class:`int`]: The minimum length of the text input."""
        return self._underlying.min_length

    @min_length.setter
    def min_length(self, value: Optional[int]) -> None:
        self._underlying.min_length = value

    @property
    def max_length(self) -> Optional[int]:
        """Optional[:class:`int`]: The maximum length of the text input."""
        return self._underlying.max_length

    @max_length.setter
    def max_length(self, value: Optional[int]) -> None:
        self._underlying.max_length = value



# SPDX-License-Identifier: MIT

from typing import Dict, Literal, Optional

import pytest

from disnake.permissions import PermissionOverwrite, Permissions


class TestPermissions:
    def test_init_permissions_keyword_arguments(self) -> None:
        perms = Permissions(manage_messages=True)

        assert perms.manage_messages is True

        # check we only have the manage message permission
        assert perms.value == Permissions.manage_messages.flag

    def test_init_permissions_keyword_arguments_with_aliases(self) -> None:
        assert Permissions(read_messages=True, view_channel=False).value == 0
        assert Permissions(view_channel=True, read_messages=False).value == 0

        assert Permissions(read_messages=False, view_channel=True).value == 1024
        assert Permissions(view_channel=False, read_messages=True).value == 1024

    def test_init_invalid_value(self) -> None:
        with pytest.raises(TypeError, match="Expected int parameter, received str instead."):
            Permissions("h")  # type: ignore

    def test_init_invalid_perms(self) -> None:
        with pytest.raises(TypeError, match="'h' is not a valid permission name."):
            Permissions(h=True)  # type: ignore

    @pytest.mark.parametrize(
        ("perms_int", "other_int", "expected"),
        [
            (0b11, 0b10, False),
            (0b10, 0b10, True),
            (0b01, 0b11, True),
        ],
    )
    def test_is_subset(self, perms_int: int, other_int: int, expected: bool) -> None:
        perms = Permissions(perms_int)
        other = Permissions(other_int)
        assert perms.is_subset(other) is expected

    def test_is_subset_only_permissions(self) -> None:
        perms = Permissions()
        with pytest.raises(TypeError, match="cannot compare Permissions with int"):
            perms.is_subset(5)  # type: ignore

    @pytest.mark.parametrize(
        ("perms_int", "other_int", "expected"),
        [
            (0b11, 0b10, True),
            (0b10, 0b10, True),
            (0b01, 0b11, False),
        ],
    )
    def test_is_superset(self, perms_int: int, other_int: int, expected: bool) -> None:
        perms = Permissions(perms_int)
        other = Permissions(other_int)
        assert perms.is_superset(other) is expected

    def test_is_superset_only_permissions(self) -> None:
        perms = Permissions()
        with pytest.raises(TypeError, match="cannot compare Permissions with int"):
            perms.is_superset(5)  # type: ignore

    @pytest.mark.parametrize(
        ("perms_int", "other_int", "expected"),
        [
            (0b11, 0b10, False),
            (0b10, 0b10, False),
            (0b01, 0b11, True),
        ],
    )
    def test_is_strict_subset(self, perms_int: int, other_int: int, expected: bool) -> None:
        perms = Permissions(perms_int)
        other = Permissions(other_int)
        assert perms.is_strict_subset(other) is expected

    @pytest.mark.parametrize(
        ("perms_int", "other_int", "expected"),
        [
            (0b11, 0b10, True),
            (0b10, 0b10, False),
            (0b01, 0b11, False),
        ],
    )
    def test_is_strict_superset(self, perms_int: int, other_int: int, expected: bool) -> None:
        perms = Permissions(perms_int)
        other = Permissions(other_int)

        assert perms.is_strict_superset(other) is expected

    @pytest.mark.parametrize(
        ("perms_dict", "update", "expected"),
        [
            (
                {"view_channel": True},
                {"move_members": True},
                {"view_channel": True, "move_members": True},
            ),
        ],
    )
    def test_update(
        self,
        perms_dict: Dict[str, bool],
        update: Dict[str, bool],
        expected: Dict[str, Literal[True]],
    ) -> None:
        perms = Permissions(**perms_dict)
        perms.update(**update)

        expected_perms = Permissions(**expected)

        assert perms.value == expected_perms.value

    @pytest.mark.parametrize(
        ("update", "expected"),
        [
            ({"read_messages": True, "view_channel": False}, 8),
            ({"view_channel": True, "read_messages": False}, 8),
            ({"read_messages": False, "view_channel": True}, 8 + 1024),
            ({"view_channel": False, "read_messages": True}, 8 + 1024),
        ],
    )
    def test_update_aliases(self, update: Dict[str, bool], expected: int) -> None:
        perms = Permissions(administrator=True)
        perms.update(**update)
        assert perms.value == expected

    @pytest.mark.parametrize(
        ("parameters", "expected"),
        [
            ({"view_channel": True, "move_members": True}, None),
            (
                # test aliases
                {"read_messages": True, "create_forum_threads": True},
                {"view_channel": True, "send_messages": True},
            ),
        ],
    )
    def test_iter(self, parameters: Dict[str, bool], expected: Optional[Dict[str, bool]]) -> None:
        perms = Permissions(**parameters)
        if expected is None:
            expected = parameters
        for key, value in iter(perms):
            assert value == expected.pop(key, False)
        assert not len(expected)

    def test_update_ignores(self) -> None:
        perms = Permissions()
        perms.update(h=True)  # type: ignore

    @pytest.mark.parametrize(
        ("initial", "allow", "deny", "expected"),
        [
            (0b1010, 0b0101, 0b1111, 0b0101),
            (0b0011, 0b0100, 0b0001, 0b0110),
            (0x0400, 0x0401, 0x5001, 0x0401),
        ],
    )
    def test_handle_overwrite(self, initial: int, allow: int, deny: int, expected: int) -> None:
        perms = Permissions(initial)
        assert perms.value == initial
        perms.handle_overwrite(allow, deny)
        assert perms.value == expected

    def test_none_is_none(self) -> None:
        perms = Permissions.none()
        assert perms.value == 0

    @pytest.mark.parametrize(
        "method_name",
        [
            "all",
            "none",
            "all_channel",
            "general",
            "membership",
            "text",
            "voice",
            "stage",
            "stage_moderator",
            "events",
            "advanced",
            "private_channel",
        ],
    )
    def test_classmethods(self, method_name: str) -> None:
        method = getattr(Permissions, method_name)

        perms: Permissions = method()
        assert isinstance(perms, Permissions)

        # check that caching does not return the same permissions instance
        perms_two: Permissions = method()
        assert perms is not perms_two
        assert perms.value == perms_two.value


class TestPermissionOverwrite:
    def test_init(self) -> None:
        perms = PermissionOverwrite(manage_messages=True)

        assert perms.manage_messages is True

    def test_init_invalid_perms(self) -> None:
        with pytest.raises(ValueError, match="'h' is not a valid permission name."):
            PermissionOverwrite(h=True)  # type: ignore

    def test_equality(self) -> None:
        one = PermissionOverwrite()
        two = PermissionOverwrite()

        assert one is not two
        assert one == two

        two.ban_members = False
        assert one != two

    def test_set(self) -> None:
        po = PermissionOverwrite()
        po.attach_files = False
        assert po.attach_files is False

        po.attach_files = True
        assert po.attach_files is True

        po.attach_files = None
        assert po.attach_files is None

    def test_set_invalid_type(self) -> None:
        po = PermissionOverwrite()
        with pytest.raises(TypeError, match="Expected bool or NoneType, received str"):
            po.connect = "h"  # type: ignore

        with pytest.raises(
            AttributeError, match="'PermissionOverwrite' object has no attribute 'oh'"
        ):
            po.oh = False  # type: ignore

    @pytest.mark.parametrize(
        ("allow", "deny"),
        [
            ({"view_channel": True}, {"ban_members": True}),
            ({"view_channel": True}, {"view_channel": True}),
            ({"administrator": True}, {"manage_channels": False}),
        ],
    )
    def test_from_pair(
        self,
        allow: Dict[str, bool],
        deny: Dict[str, bool],
    ) -> None:
        perm_allow = Permissions(**allow)
        perm_deny = Permissions(**deny)

        po = PermissionOverwrite.from_pair(perm_allow, perm_deny)

        # iterate over the allowed perms and assert that the overwrite is what the allowed perms are
        for perm, allowed in perm_allow:
            # get the attr from the denied perms as denied perms override the allow list in from_pair
            if allowed and not getattr(perm_deny, perm):
                assert getattr(po, perm) is True
            else:
                assert getattr(po, perm) is not True

        for perm, denied in deny.items():
            if denied:
                assert getattr(po, perm) is False
            else:
                assert getattr(po, perm) is not False

    @pytest.mark.parametrize(
        ("allow", "deny"),
        [
            # these intentionally do not interfere with each other
            (0b1, 0b10),
            (0x313, 0x424),
            (0x69420, 0x2301),
        ],
    )
    def test_pair(self, allow: int, deny: int) -> None:
        og_perms_allow = Permissions(allow)
        og_perms_deny = Permissions(deny)

        po = PermissionOverwrite.from_pair(og_perms_allow, og_perms_deny)

        perms_allow, perms_deny = po.pair()

        assert perms_allow.value == og_perms_allow.value
        assert perms_deny.value == og_perms_deny.value

    def test_is_empty(self) -> None:
        po = PermissionOverwrite()
        assert po.is_empty()

        po.add_reactions = True
        assert not po.is_empty()

    def test_update(self) -> None:
        po = PermissionOverwrite()
        assert po.manage_emojis is None

        po.update(manage_emojis=True)
        assert po.manage_emojis is True

        assert po.manage_permissions is None
        po.update(manage_permissions=False)
        assert po.manage_permissions is False

        po.update(manage_permissions=None, manage_emojis=None)
        assert po.manage_permissions is None
        assert po.manage_emojis is None

        # invalid names are silently ignored
        po.update(h=True)  # type: ignore
        assert not hasattr(po, "h")

    @pytest.mark.parametrize(
        ("expected"),
        [
            ({"view_channel": True}),
            ({"ban_members": None}),
            ({"view_channel": True, "administrator": False, "ban_members": None}),
            ({"kick_members": False}),
        ],
    )
    def test_iter(
        self,
        expected: Dict[str, bool],
    ) -> None:
        po = PermissionOverwrite(**expected)

        for perm, value in po:
            assert expected.get(perm, None) is value



# SPDX-License-Identifier: MIT

from typing import Literal, TypedDict

from typing_extensions import NotRequired

from .i18n import LocalizationDict

ApplicationRoleConnectionMetadataType = Literal[1, 2, 3, 4, 5, 6, 7, 8]


class ApplicationRoleConnectionMetadata(TypedDict):
    type: ApplicationRoleConnectionMetadataType
    key: str
    name: str
    name_localizations: NotRequired[LocalizationDict]
    description: str
    description_localizations: NotRequired[LocalizationDict]
